import asyncio
import uuid
from functools import partial
from unittest.mock import Mock

import pytest

from cashews.backends.memory import Memory
from cashews.decorators.rate import RateLimitException
from cashews.decorators.rate import rate_limit as _rate_limit

pytestmark = pytest.mark.asyncio


@pytest.fixture()
async def rate_limit():
    return partial(_rate_limit, backend=Memory())


@pytest.mark.parametrize("n", list(range(1, 10)))
async def test_rate_limit_simple(rate_limit, n):
    @rate_limit(limit=n, period=0.01, prefix=str(uuid.uuid4()))
    async def func():
        return 1

    for i in range(n):
        assert await func() == 1

    with pytest.raises(RateLimitException):
        await func()

    await asyncio.sleep(0.01)
    assert await func() == 1


async def test_rate_limit_ttl(rate_limit):
    @rate_limit(limit=1, period=0.02)
    async def func():
        return 1

    assert await func() == 1

    with pytest.raises(RateLimitException):
        await func()

    await asyncio.sleep(0.01)
    with pytest.raises(RateLimitException):
        await func()

    await asyncio.sleep(0.01)
    await func()


async def test_rate_limit_func_args_dict(rate_limit):
    @rate_limit(limit=1, period=0.1, key="user:{user.name}")
    async def func(user):
        return user.name

    obj = type("user", (), {"name": "test"})()

    assert await func(obj) == "test"

    with pytest.raises(RateLimitException):
        await func(obj)

    obj = type("user", (), {"name": "new"})()
    assert await func(obj) == "new"


async def test_rate_limit_action(rate_limit):
    action = Mock()

    @rate_limit(limit=1, period=0.1, action=action, key="key")
    async def func(k=None):
        return 1

    assert await func() == 1
    action.assert_not_called()

    assert await func(k="test") == 1
    action.assert_called_with(k="test")



import random
from contextlib import nullcontext

import pytest
from _pytest.fixtures import SubRequest

from cashews.serialize import PickleSerializerMixin, SignIsMissingError, UnSecureDataError


class TestPickleSerializerMixin:
    @pytest.fixture(params=[None, "sample_sign"])
    def pickle_serializer_mixin(self, request: SubRequest) -> PickleSerializerMixin:
        return PickleSerializerMixin(
            hash_key=request.param,
            check_repr=random.choice([True, False]),
        )

    @pytest.mark.parametrize(
        ("key", "value", "value_has_empty_sign", "value_has_valid_sign"),
        [
            ("without sign and without underscore separator", b"spam eggs", False, True),
            (
                "with sign as an empty string and with underscore separator",
                b"_spam eggs",  # Check backward compatibility: when `hash_key` was not used.
                True,
                False,
            ),
        ],
    )
    def test__split_value_from_signature(
        self,
        pickle_serializer_mixin: PickleSerializerMixin,
        key: str,
        value: bytes,
        value_has_empty_sign: bool,
        value_has_valid_sign: bool,
    ) -> None:
        ctx = nullcontext()
        if pickle_serializer_mixin._hash_key:
            if value_has_empty_sign and not value_has_valid_sign:
                # If we use a `hash_key`, but a value is signed with an empty or a wrong string, then we expect `UnSecureDataError`.
                ctx = pytest.raises(UnSecureDataError)
            elif not value_has_empty_sign:
                # If we use a `hash_key`, but a value is not signed, then we expect `SignIsMissingError`.
                ctx = pytest.raises(SignIsMissingError, match=key)

        with ctx:
            assert pickle_serializer_mixin._split_value_from_signature(value, key) == b"spam eggs"



import base64
import json
import re
from hashlib import md5, sha1, sha256
from string import Formatter
from typing import Any, Callable, Iterable, Optional, Tuple


def _decode_bytes(value: bytes):
    try:
        return value.decode()
    except UnicodeDecodeError:
        return value.hex()


def _get_decode_array(format_value):
    def _decode_array(values: Iterable[str]) -> str:
        return ":".join([format_value(value) for value in values])

    return _decode_array


def _get_decoded_dict(format_value):
    def _decode_dict(value: dict):
        return ":".join(f"{k}:{format_value(v)}" for k, v in sorted(value.items()))

    return _decode_dict


class _ReplaceFormatter(Formatter):
    def __init__(self, default=lambda field: "*"):
        self.__default = default
        _decode_array = _get_decode_array(self._format_field)
        self.__type_format = {
            bool: lambda value: str(value).lower(),
            bytes: _decode_bytes,
            tuple: _decode_array,
            list: _decode_array,
            set: _decode_array,
            dict: _get_decoded_dict(self._format_field),
        }
        super().__init__()

    def set_format_for_type(self, type, format_function):
        self.__type_format[type] = format_function

    def type_format(self, type):
        def _decorator(func):
            self.set_format_for_type(type, func)
            return func

        return _decorator

    def get_field(self, field_name, args, kwargs):
        try:
            return super().get_field(field_name, args, kwargs)
        except (KeyError, AttributeError):
            return self.__default(field_name), None

    def _format_field(self, value):
        if value is None:
            return ""
        _type = type(value)
        if _type in self.__type_format:
            return self.__type_format[_type](value)
        for _type, func_format in self.__type_format.items():
            if isinstance(value, _type):
                return func_format(value)
        return str(value)

    def format_field(self, value, format_spec):
        return format(self._format_field(value))


class _FuncFormatter(_ReplaceFormatter):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._functions = {}
        super().__init__(*args, **kwargs)

    def _register(self, alias, function) -> None:
        self._functions[alias] = function

    def register(self, alias):
        def _decorator(func):
            self._register(alias, func)
            return func

        return _decorator

    def format_field(self, value, format_spec):
        format_spec, args = self.parse_format_spec(format_spec)
        if format_spec in self._functions:
            value = self._functions[format_spec](value, *args)
        value = super().format_field(value, format_spec if format_spec not in self._functions else "")
        return value

    @staticmethod
    def parse_format_spec(format_spec):
        if not format_spec or "(" not in format_spec:
            return format_spec, ()
        format_spec, args = format_spec.split("(", 1)
        return format_spec, args.replace(")", "").split(",")


default_formatter = _FuncFormatter(lambda name: "")
default_formatter._register("len", len)


@default_formatter.register("jwt")
def _jwt_func(jwt: str, key: str):
    _, payload, _ = jwt.split(".", 2)
    payload_dict = json.loads(base64.b64decode(payload))
    return payload_dict.get(key)


@default_formatter.register("hash")
def _hash_func(value: str, alg="md5") -> str:
    algs = {"sha1": sha1, "md5": md5, "sha256": sha256}
    alg = algs[alg]
    return alg(value.encode()).hexdigest()


def template_to_pattern(template: str, _formatter=_ReplaceFormatter(), **values) -> str:
    return _formatter.format(template, **values)


def _re_default(field_name):
    return f"(?P<{field_name.replace('.', '_')}>[^:]*)"


_re_formatter = _ReplaceFormatter(default=_re_default)
_REGISTER = {}


def register_template(func, template: str):
    pattern = "(.*[:])?" + template_to_pattern(template, _formatter=_re_formatter) + "$"
    compile_pattern = re.compile(pattern, flags=re.MULTILINE)
    _REGISTER.setdefault((func.__module__ or "", func.__name__), set()).add((template, compile_pattern))


def get_templates_for_func(func):
    return (template for template, _ in _REGISTER.get((func.__module__ or "", func.__name__), set()))


def get_template_and_func_for(key: str) -> Tuple[Optional[str], Optional[Callable]]:
    for func, templates in _REGISTER.items():
        for template, compile_pattern in templates:
            if compile_pattern.fullmatch(key):
                return template_to_pattern(template), func
    return None, None


def get_template_for_key(key: str) -> Tuple[Optional[str], Optional[dict]]:
    for func, templates in _REGISTER.items():
        for template, compile_pattern in templates:
            match = compile_pattern.fullmatch(key)
            if match:
                return template, match.groupdict()
    return None, None



import asyncio
import logging
from datetime import datetime, timedelta
from functools import wraps
from typing import Optional

from ..._typing import CacheCondition
from ...backends.interface import Backend
from ...formatter import register_template
from ...key import get_cache_key, get_cache_key_template
from .defaults import CacheDetect, _empty, _get_cache_condition, context_cache_detect

__all__ = ("early",)
logger = logging.getLogger(__name__)
_LOCK_SUFFIX = ":lock"


def early(
    backend: Backend,
    ttl: int,
    key: Optional[str] = None,
    early_ttl: Optional[int] = None,
    condition: CacheCondition = None,
    prefix: str = "early",
):
    """
    Cache strategy that try to solve Cache stampede problem (https://en.wikipedia.org/wiki/Cache_stampede),
    With hot cache recalculate a result in background near expiration time
    Warning Not good at cold cache

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    """
    store = _get_cache_condition(condition)
    if early_ttl is None:
        early_ttl = ttl * 0.33

    def _decor(func):
        _key_template = get_cache_key_template(func, key=key, prefix=prefix + ":v2")
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                _from_cache._set(
                    _cache_key,
                    ttl=ttl,
                    early_ttl=early_ttl,
                    name="early",
                    template=_key_template,
                )
                early_expire_at, result = cached
                if early_expire_at <= datetime.utcnow() and await backend.set(
                    _cache_key + _LOCK_SUFFIX, "1", expire=early_ttl, exist=False
                ):
                    logger.info(
                        "Recalculate cache for %s (exp_at %s)",
                        _cache_key,
                        early_expire_at,
                    )
                    asyncio.create_task(
                        _get_result_for_early(
                            backend,
                            func,
                            args,
                            kwargs,
                            _cache_key,
                            ttl,
                            early_ttl,
                            store,
                        )
                    )
                return result
            return await _get_result_for_early(
                backend, func, args, kwargs, _cache_key, ttl, early_ttl, store, unlock=True
            )

        return _wrap

    return _decor


async def _get_result_for_early(
    backend: Backend, func, args, kwargs, key, ttl: int, early_ttl: int, condition, unlock=False
):
    try:
        result = await func(*args, **kwargs)
        if condition(result, args, kwargs, key):
            early_expire_at = datetime.utcnow() + timedelta(seconds=early_ttl)
            await backend.set(key, [early_expire_at, result], expire=ttl)
        return result
    finally:
        if unlock:
            asyncio.create_task(backend.delete(key + _LOCK_SUFFIX))



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Type

import pytest

if TYPE_CHECKING:
    from cashews.backends.interface import Backend


pytest_plugins = ["pytest_asyncio"]  # pylint: disable=invalid-name

try:
    import aiohttp
except ImportError:
    pass
else:
    pytest_plugins.append("aiohttp.pytest_plugin")

    del aiohttp


@pytest.fixture
def redis_dsn():
    host = os.getenv("REDIS_HOST", "")
    port = os.getenv("REDIS_PORT", "")
    return f"redis://{host}:{port}/"


@pytest.fixture
async def backend_factory():
    backend = None

    async def factory(backend_cls: Type[Backend], *args, **kwargs):
        nonlocal backend

        backend = backend_cls(*args, **kwargs)
        await backend.init()
        await backend.clear()

        return backend

    try:
        yield factory
    finally:
        assert backend is not None, "Fixture `backend_factory` wasn't called."
        backend.close()



import uuid
from contextlib import asynccontextmanager
from typing import Any, AsyncIterator, Optional, Tuple


class LockedException(Exception):
    pass


class Backend:
    name: str = ""
    is_init: bool = False

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    async def init(self):
        ...

    def close(self):
        ...

    async def set(
        self,
        key: str,
        value: Any,
        expire: Optional[float] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        ...

    async def set_raw(self, key: str, value: Any, **kwargs: Any):
        ...

    async def get(self, key: str, default: Optional[Any] = None) -> Any:
        ...

    async def get_raw(self, key: str) -> Any:
        ...

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any, ...]:
        ...

    async def exists(self, key: str) -> bool:
        ...

    async def keys_match(self, pattern: str) -> AsyncIterator[str]:
        ...

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:
        ...

    async def incr(self, key: str) -> int:
        ...

    async def delete(self, key: str) -> bool:
        ...

    async def delete_match(self, pattern: str):
        ...

    async def get_match(
        self, pattern: str, batch_size: int = 100, default: Optional[Any] = None
    ) -> AsyncIterator[Tuple[str, Any]]:
        ...

    async def expire(self, key: str, timeout: float):
        ...

    async def get_expire(self, key: str) -> int:
        ...

    async def get_bits(self, key: str, *indexes: int, size: int = 1) -> Tuple[int]:
        ...

    async def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1) -> Tuple[int]:
        ...

    async def get_size(self, key: str) -> int:
        """
        Return size in bites that allocated by a value for given key
        """
        ...

    async def ping(self, message: Optional[bytes] = None) -> bytes:
        ...

    async def clear(self):
        ...

    async def set_lock(self, key: str, value: Any, expire: float) -> bool:
        ...

    async def is_locked(
        self,
        key: str,
        wait: Optional[float] = None,
        step: float = 0.1,
    ) -> bool:
        ...

    async def unlock(self, key: str, value: Any) -> bool:
        ...

    @asynccontextmanager
    async def lock(self, key: str, expire: float):
        identifier = str(uuid.uuid4())
        lock = await self.set_lock(key, identifier, expire=expire)
        if not lock:
            # we need to check the connection by ping because
            # if for example a redis unavailable and a backend have flag `safe`
            # we will have a brake lock
            try:
                await self.ping(b"TEST")
            except Exception:
                pass
            else:
                raise LockedException(f"Key {key} already locked")
        try:
            yield
        finally:
            await self.unlock(key, identifier)



"""
Client side cache is hybrid of mem and redis cache
in theory local cache should be consistence
GET:
-> IN mem cache -> Y -> return
                -> N -> in redis cache -> Y -> store in mem cache -> return
                                       -> N -> compete -> store in mem and in redis -> notify others by channel to invalidete

INVALIDATE:

problem with redis client side cache  - if client set he didnt receive message (so we can't update local cache on set without get on redis)
                                        - message only for first set (2 and after will miss) (solve by request resource after message
                                        - no control
                                        - redis >= 6
                                        + mem cache without ttl
                                        + no trash

Redis client side caching with non broadcast option weed, with pool of connections it is hard to process connection
lifetime and subscribe for get requests also if we set some value with ttl every client who get value from redis can
also request a ttl and store in local cache with ttl but we steal should know if someone overwrite value or delete it
Broadcasting mode is more useful as we can subscribe for all keys with prefix and invalidate key

https://engineering.redislabs.com/posts/redis-assisted-client-side-caching-in-python/
https://redis.io/topics/client-side-caching
"""

import asyncio
import logging
from typing import Any, AsyncIterator, Optional, Tuple

try:
    from redis.exceptions import ConnectionError as RedisConnectionError
except ImportError:
    from aioredis import RedisError as RedisConnectionError

from .memory import Memory
from .redis import Redis

_REDIS_INVALIDATE_CHAN = "__redis__:invalidate"
_empty = object()
_RECONNECT_WAIT = 10
_DEFAULT_PREFIX = "cashews:"
BCAST_ON = "CLIENT TRACKING on REDIRECT {client_id} BCAST PREFIX {prefix}"
logger = logging.getLogger(__name__)


class BcastClientSide(Redis):
    """
    Cache backend with redis as main storage and client side mem storage that invalidated by
    redis channel for client-side-caching.

    Subscribe with broadcasting by prefix for invalidate by redis>=6
    https://redis.io/topics/client-side-caching
    """

    name = "redis_mem"

    def __init__(self, *args: Any, local_cache=None, client_side_prefix: str = _DEFAULT_PREFIX, **kwargs: Any) -> None:
        self._local_cache = Memory() if local_cache is None else local_cache
        self._prefix = client_side_prefix
        self._recently_update = Memory(size=500, check_interval=5)
        self.__listen_task = None
        self._listen_started = None
        super().__init__(*args, **kwargs)

    async def init(self):
        self._listen_started = asyncio.Event()
        await self._local_cache.init()
        await self._recently_update.init()
        await super().init()
        self.__listen_task = asyncio.create_task(self._listen_invalidate_forever())
        await self._listen_started.wait()

    async def _mark_as_recently_updated(self, key: str):
        await self._recently_update.set(key, True, expire=5)

    async def _get_local_value(self, key: str):
        if not self._listen_started.is_set():
            return _empty
        return await self._local_cache.get(key, default=_empty)

    async def _listen_invalidate_forever(self):
        while True:
            try:
                await self._listen_invalidate()
            except (RedisConnectionError, ConnectionRefusedError):
                logger.error("broken connection with redis. Clearing client side storage")
                self._listen_started = asyncio.Event()
                await self._local_cache.clear()
                await asyncio.sleep(_RECONNECT_WAIT)

    async def _get_channel(self):
        pubsub = self._client.pubsub()
        await pubsub.execute_command(b"CLIENT", b"ID")
        client_id = await pubsub.parse_response()
        await pubsub.execute_command(*BCAST_ON.format(client_id=client_id, prefix=self._prefix).encode().split())
        await pubsub.parse_response()
        await pubsub.subscribe(_REDIS_INVALIDATE_CHAN)
        return pubsub

    async def _listen_invalidate(self):
        channel = await self._get_channel()
        self._listen_started.set()
        await self._local_cache.clear()
        while True:
            message = await channel.get_message(ignore_subscribe_messages=True, timeout=0.1)
            if message is None or not message.get("data"):
                continue
            key = message["data"][0]
            key = key.decode().lstrip(self._prefix)
            if not await self._recently_update.get(key):
                await self._local_cache.delete(key)
            else:
                await self._recently_update.delete(key)

    async def get(self, key: str, default: Any = None) -> Any:
        value = await self._get_local_value(key)
        if value is not _empty:
            return value
        value = await super().get(self._prefix + key, default=_empty)
        if value is not _empty:
            await self._local_cache.set(key, value)
            return value
        return default

    async def set(self, key: str, value: Any, *args: Any, **kwargs: Any) -> Any:
        if await self._local_cache.get(key, default=_empty) == value:
            # If value in current client_cache - skip resetting
            return 0
        await self._local_cache.set(key, value, *args, **kwargs)  # not async by the way
        await self._mark_as_recently_updated(key)
        return await super().set(self._prefix + key, value, *args, **kwargs)

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:
        async for key in super().scan(self._prefix + pattern, batch_size=batch_size):
            yield key.strip(self._prefix)

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any]:
        missed_keys = {self._prefix + key for key in keys}
        values = {key: default for key in keys}
        if self._listen_started.is_set():
            for i, value in enumerate(await self._local_cache.get_many(*keys, default=_empty)):
                key = keys[i]
                if value is not _empty:
                    values[key] = value
                    missed_keys.remove(self._prefix + key)
        missed_values = await super().get_many(*missed_keys, default=default)
        missed = dict(zip((key.strip(self._prefix) for key in missed_keys), missed_values))
        for key, value in missed.items():
            if value is not default:
                await self._local_cache.set(key, value)
        return tuple(missed.get(key, value) for key, value in values.items())

    async def get_match(
        self, pattern: str, batch_size: int = 100, default: Optional[Any] = None
    ) -> AsyncIterator[Tuple[str, bytes]]:
        keys_in_local_cache = set()
        async for key, value in self._local_cache.get_match(pattern, batch_size=batch_size, default=_empty):
            keys_in_local_cache.add(key)
            yield key, value
        cursor = b"0"
        while cursor:
            cursor, keys = await self._client.scan(cursor, match=self._prefix + pattern, count=batch_size)
            if not keys:
                continue
            keys = [key.decode() for key in keys]
            keys = [key for key in keys if key.strip(self._prefix) not in keys_in_local_cache]
            values = await super().get_many(*keys, default=default)
            for key, value in zip(keys, values):
                yield key.strip(self._prefix), value

    async def incr(self, key: str) -> int:
        await self._local_cache.incr(key)
        await self._mark_as_recently_updated(key)
        return await super().incr(self._prefix + key)

    async def delete(self, key: str) -> bool:
        await self._local_cache.delete(key)
        return await super().delete(self._prefix + key)

    async def delete_match(self, pattern: str):
        await self._local_cache.delete_match(pattern)
        return await super().delete_match(self._prefix + pattern)

    async def expire(self, key: str, timeout: Optional[float]):
        local_value = await self._local_cache.get(key, default=_empty)
        if local_value is not _empty:
            await self._local_cache.expire(key, timeout)
            await self._mark_as_recently_updated(key)
        result = await super().expire(self._prefix + key, timeout)
        return result

    async def get_expire(self, key: str) -> int:
        if await self._local_cache.get_expire(key) != -1:
            if self._listen_started.is_set():
                return await self._local_cache.get_expire(key)
        expire = await super().get_expire(self._prefix + key)
        await self._local_cache.expire(key, expire)
        return expire

    async def exists(self, key) -> bool:
        if self._listen_started.is_set():
            if await self._local_cache.exists(key):
                return True
        return await super().exists(self._prefix + key)

    async def set_lock(self, key: str, value, expire):
        await self._mark_as_recently_updated(key)
        await self._local_cache.set_lock(key, value, expire)
        pexpire = None
        if isinstance(expire, float):
            pexpire = int(expire * 1000)
            expire = None
        return bool(await self._client.set(self._prefix + key, value, ex=expire, px=pexpire, nx=True))

    async def unlock(self, key, value):
        await self._local_cache.unlock(key, value)
        return await super().unlock(self._prefix + key, value)

    async def get_size(self, key: str) -> int:
        return await super().get_size(self._prefix + key)

    async def clear(self):
        await self._local_cache.clear()
        return await super().clear()

    def close(self):
        if self.__listen_task is not None:
            self.__listen_task.cancel()
            self.__listen_task = None
        self._local_cache.close()
        self._recently_update.close()
        super().close()



from cashews.utils import Bitarray, get_hashes


def test_bitarray_get_size_1():
    b = Bitarray("100010", 2)
    assert b.get(0) == 0
    assert b.get(1) == 1
    assert b.get(2) == 0
    assert b.get(3) == 0
    assert b.get(4) == 0
    assert b.get(5) == 1
    assert b.get(6) == 0
    assert b.get(7) == 0


def test_bitarray_get_size_2():
    b = Bitarray("1000010", 2)
    assert b.get(0, 2) == 2
    assert b.get(1, 2) == 0
    assert b.get(2, 2) == 0
    assert b.get(3, 2) == 1
    assert b.get(4, 2) == 0


def test_bitarray_get_size_3():
    b = Bitarray("100010", 2)
    assert b.get(0, 3) == 2
    assert b.get(1, 3) == 4


def test_bitarray_incr_size_1():
    b = Bitarray("100010", 2)
    assert str(b.copy().incr(0)) == "100011"
    assert str(b.copy().incr(0, by=-1)) == "100010"
    assert str(b.copy().incr(1)) == "100010"
    assert str(b.copy().incr(1, by=-1)) == "100000"
    assert str(b.copy().incr(2)) == "100110"


def test_bitarray_incr_size_2():
    b = Bitarray("100010", 2)
    assert str(b.copy().incr(0, size=2)) == "100011"
    assert str(b.copy().incr(0, size=2, by=2)) == "100011"
    assert str(b.copy().incr(0, size=2, by=3)) == "100011"

    assert str(b.copy().incr(1, size=2)) == "100110"
    assert str(b.copy().incr(1, size=2, by=3)) == "101110"
    assert str(b.copy().incr(1, size=2, by=10)) == "101110"
    assert str(b.copy().incr(1, size=2, by=-1)) == "100010"


def test_bitarray_incr_size_3():
    b = Bitarray("100010", 2)
    assert str(b.copy().incr(0, size=3)) == "100011"
    assert str(b.copy().incr(0, size=3, by=2)) == "100100"
    assert str(b.copy().incr(0, size=3, by=3)) == "100101"
    assert str(b.copy().incr(0, size=3, by=6)) == "100111"


def test_bitarray_incr_over():
    b = Bitarray("1", 2)
    assert str(b.incr(2)) == "101"


def test_get_hashes_params():
    assert len(get_hashes("test", 3, 3)) == 3
    assert len(get_hashes("test", 1, 100)) == 1
    assert len(get_hashes("test", 77, 100)) == 77
    assert len(get_hashes("long string hash", 77, 100)) == 77
    assert get_hashes("test", 1, 100).pop() < 100


def test_get_hashes():
    assert get_hashes("test", 10, 1000) == get_hashes("test", 10, 1000)
    assert get_hashes("test", 2, 2) == get_hashes("test", 2, 2)

    assert get_hashes("test", 2, 5) != get_hashes("test", 2, 2)
    assert get_hashes("test", 3, 100) != get_hashes("tset", 3, 100)
    assert get_hashes("test", 3, 100) != get_hashes("t", 3, 100)
    assert get_hashes("test", 3, 100) != get_hashes("tes", 3, 100)
    assert get_hashes("test", 3, 100) != get_hashes("test2", 3, 100)
    assert not get_hashes("test", 5, 100).intersection(get_hashes("a", 5, 100))
    assert not get_hashes("test", 5, 100).intersection(get_hashes("test2", 5, 100))

    assert get_hashes("test", 3, 3) == get_hashes("a", 3, 3)  # it is ok to have collisions in this case
    assert len(get_hashes("test", 20, 100)) == len(set(get_hashes("test", 20, 100)))



from bitarray import bitarray, util


class Bitarray:
    __slots__ = ("_value",)

    def __init__(self, value, base=10):
        i = int(value, base=base)
        self._value = bitarray(f"{i:b}")

    def get(self, index: int, size: int = 1) -> int:
        start = max(len(self._value) - (index + 1) * size, 0)
        finish = max(len(self._value) - index * size, 0)
        bits = self._value[start:finish]
        if len(bits) == 0:
            return 0
        return util.ba2int(bits)

    def set(self, index: int, value: int, size: int = 1):
        for i in range(0, size):
            index_to_set = index * size + i
            if len(self._value) - 1 < index_to_set:
                self._value = bitarray((index_to_set - len(self._value) + 1) * "0" + self._value.to01())
            self._value[len(self._value) - index_to_set - 1] = (value >> i) & 1
        return

    def _set_bit_1(self, index):
        self._value[len(self._value) - index] = 1

    def _set_bit_0(self, index):
        self._value[len(self._value) - index] = 0

    def incr(self, index: int, size: int = 1, by: int = 1):
        if by > 0:
            by = min(by, 2**size - 1)
        else:
            by = max(by, -(2**size) - 1)
        value = self.get(index, size)
        value += by
        value = min(max(0, value), 2**size - 1)
        self.set(index, value, size)
        return self

    def copy(self):
        return Bitarray(self._value.to01(), base=2)

    def to_int(self):
        return util.ba2int(self._value, False)

    def __str__(self):
        return self._value.to01()



import asyncio
from unittest.mock import Mock

import pytest

from cashews import NOT_NONE, decorators, noself
from cashews.backends.memory import Backend, Memory
from cashews.formatter import _REGISTER, get_templates_for_func

pytestmark = pytest.mark.asyncio

EXPIRE = 0.02


class CustomError(Exception):
    pass


@pytest.fixture(
    name="backend",
    params=[
        "memory",
        pytest.param("redis", marks=pytest.mark.redis),
        # pytest.param("redis_cs", marks=pytest.mark.redis),
        pytest.param("diskcache", marks=pytest.mark.diskcache),
    ],
)
async def _backend(request, redis_dsn, backend_factory):
    if request.param == "diskcache":
        from cashews.backends.diskcache import DiskCache

        yield await backend_factory(DiskCache)
    elif request.param == "redis":
        from cashews.backends.redis import Redis

        yield await backend_factory(Redis, redis_dsn, max_connections=20)
    else:
        yield await backend_factory(backend_cls=Memory)


async def test_fail_cache_simple(backend):
    @decorators.failover(backend, ttl=EXPIRE, exceptions=CustomError, key="fail")
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    assert await func() == b"ok"
    assert await func(fail=True) == b"ok"

    await asyncio.sleep(EXPIRE * 1.1)
    with pytest.raises(CustomError):
        await func(fail=True)


async def test_circuit_breaker_simple(backend):
    @decorators.circuit_breaker(backend, ttl=EXPIRE * 10, min_calls=10, errors_rate=5, period=1, key="test")
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    for _ in range(9):
        assert await func() == b"ok"

    with pytest.raises(CustomError):
        await func(fail=True)
    await asyncio.sleep(0)

    with pytest.raises(decorators.CircuitBreakerOpen):
        await func(fail=True)

    with pytest.raises(decorators.CircuitBreakerOpen):
        await func(fail=False)


async def test_circuit_breaker_half_open(backend):
    @decorators.circuit_breaker(
        backend,
        ttl=EXPIRE,
        half_open_ttl=0.1,
        errors_rate=1,
        min_calls=0,
        period=1,
        key="test",
    )
    async def func(fail=False):
        if fail:
            raise CustomError()
        return b"ok"

    with pytest.raises(CustomError):
        await func(fail=True)

    with pytest.raises(decorators.CircuitBreakerOpen):
        await func(fail=True)

    await asyncio.sleep(EXPIRE)

    errors = 0
    success = 0
    for _ in range(100):
        try:
            await func(fail=False)
        except decorators.CircuitBreakerOpen:
            errors += 1
        else:
            success += 1
    assert success > 0
    assert errors > 0
    assert success + errors == 100

    await asyncio.sleep(0.1)
    assert await func() == b"ok"


async def test_cache_simple(backend):
    @decorators.cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    assert await func() == b"ok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"


async def test_cache_simple_not_none(backend):
    mock = Mock()

    @decorators.cache(backend, ttl=EXPIRE, key="key", condition=NOT_NONE)
    async def func():
        mock()
        return None

    assert await func() is None
    assert mock.call_count == 1

    assert await func() is None
    assert mock.call_count == 2


async def test_cache_simple_none(backend):
    mock = Mock()

    @decorators.cache(backend, ttl=EXPIRE, key="key")
    async def func():
        mock()
        return None

    assert await func() is None
    assert mock.call_count == 1

    assert await func() is None
    assert mock.call_count == 1


async def test_cache_simple_key(backend):
    _REGISTER.clear()

    @decorators.cache(backend, ttl=1, key="key:{some}")
    async def func1(resp=b"ok", some="err"):
        return resp

    @decorators.cache(backend, ttl=1)
    async def func2(resp, some="err"):
        return resp

    _noself = noself(decorators.cache)(backend, ttl=1)

    class Klass:
        @decorators.cache(backend, ttl=1)
        async def method(self, resp):
            return resp

        @staticmethod
        @decorators.cache(backend, ttl=1)
        async def stat(resp):
            return resp

        @_noself
        async def method2(self, resp):
            return resp

    await func1()
    await func2("ok")
    obj = Klass()
    await obj.method("ok")
    await obj.stat("ok")
    await obj.method2("ok")

    assert next(get_templates_for_func(func1)) == "key:{some}"
    assert next(get_templates_for_func(func2)) == "tests.test_cache:func2:resp:{resp}:some:{some}"
    assert (
        next(get_templates_for_func(obj.method))
        == "tests.test_cache:test_cache_simple_key.<locals>.Klass.method:self:{self}:resp:{resp}"
    )
    assert (
        next(get_templates_for_func(obj.method2))
        == "tests.test_cache:test_cache_simple_key.<locals>.Klass.method2:resp:{resp}"
    )
    assert next(get_templates_for_func(obj.stat)) == "tests.test_cache:stat:resp:{resp}"


async def test_cache_simple_cond(backend):
    mock = Mock()

    @decorators.cache(backend, ttl=EXPIRE, key="key", condition=lambda x, *args, **kwargs: x == b"hit")
    async def func(resp=b"ok"):
        mock()
        return resp

    await func()
    assert mock.call_count == 1

    await func(b"notok")
    assert mock.call_count == 2

    await func(b"hit")
    assert mock.call_count == 3
    await func(b"hit")
    assert mock.call_count == 3


async def test_early_cache_simple(backend):
    @decorators.early(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    assert await func() == b"ok"

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"


async def test_soft_cache_simple(backend):
    mock = Mock()

    @decorators.soft(backend, ttl=4 * EXPIRE, soft_ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        mock()
        return resp

    assert await func() == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 1.1)
    assert await func(b"notok") == b"notok"
    assert mock.call_count == 2


async def test_soft_cache_on_exc(backend):
    mock = Mock()

    @decorators.soft(backend, ttl=4 * EXPIRE, soft_ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        if resp == b"notok":
            raise ValueError()
        mock()
        return resp

    assert await func() == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(0)
    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 1.1)

    assert await func(b"notok") == b"ok"
    assert mock.call_count == 1

    await asyncio.sleep(EXPIRE * 3)
    with pytest.raises(ValueError):
        await func(b"notok")


async def test_early_cache_parallel(backend):
    if backend.name in ("redis", "diskcache"):
        pytest.skip("fail in ci with slow redis or disk")

    mock = Mock()

    @decorators.early(backend, ttl=0.1, early_ttl=0.05, key="key")
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    assert await func() == b"ok"  # warm

    assert mock.call_count == 1

    for _ in range(8):  # 8 * 0.01 = 0.08 - 1 more time should be executed
        await asyncio.sleep(0.01)
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count == 2


async def test_lock_cache_parallel(backend):
    mock = Mock()

    @decorators.locked(backend, key="key", step=0.01)
    async def func():
        await asyncio.sleep(0.1)
        mock()

    for _ in range(2):
        await asyncio.gather(*[func() for _ in range(10)], return_exceptions=True)

    assert mock.call_count == 2


async def test_lock_cache_parallel_ttl(backend):
    mock = Mock()

    @decorators.locked(backend, key="key", step=0.01, ttl=0.1)
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    for _ in range(4):
        await asyncio.gather(*[func() for _ in range(10)], return_exceptions=True)

    assert mock.call_count == 40


async def test_lock_cache_broken_backend():
    backend = Mock(wraps=Backend())
    mock = Mock()

    @decorators.locked(backend, key="key", step=0.01)
    async def func(resp=b"ok"):
        await asyncio.sleep(0.01)
        mock(resp)
        return resp

    for _ in range(4):
        await asyncio.gather(*[func() for _ in range(10)])

    assert mock.call_count == 40


async def test_hit_cache(backend):
    mock = Mock()

    @decorators.hit(backend, ttl=10, cache_hits=10, key="test")
    async def func(resp=b"ok"):
        mock(resp)
        return resp

    await func()  # cache
    await asyncio.gather(*[func() for _ in range(10)])  # get 10 hits
    assert mock.call_count == 1
    await func()  # cache
    assert mock.call_count == 2

    await asyncio.gather(*[func() for _ in range(10)])
    assert mock.call_count == 2


async def test_hit_cache_early(backend):
    mock = Mock()

    @decorators.hit(backend, ttl=10, cache_hits=5, key="test", update_after=1)
    async def func(resp=b"ok"):
        mock(resp)
        return resp

    assert await func(b"1") == b"1"  # nocache
    assert mock.call_count == 1

    assert await func(b"2") == b"1"  # cache and update
    assert mock.call_count == 1

    await asyncio.sleep(0.01)
    assert mock.call_count == 2
    assert await func(b"3") == b"2"  # cache from prev and also update
    await asyncio.sleep(0.01)
    assert mock.call_count == 3


async def test_cache_detect_simple(backend):
    @decorators.cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    cache_detect = decorators.CacheDetect()
    assert await func(_from_cache=cache_detect) == b"ok"
    assert cache_detect.keys == {}

    await asyncio.sleep(0)
    assert await func(b"notok", _from_cache=cache_detect) == b"ok"
    assert len(cache_detect.keys) == 1
    assert list(cache_detect.keys.keys()) == [
        "key",
    ]


async def test_context_cache_detect_simple(backend):
    @decorators.cache(backend, ttl=EXPIRE, key="key")
    async def func(resp=b"ok"):
        return resp

    with decorators.context_cache_detect as detector:
        assert await func() == b"ok"
        assert detector.keys == {}

        await asyncio.sleep(0)
        assert await func(b"notok") == b"ok"
        assert list(detector.keys.keys()) == [
            "key",
        ]

        await asyncio.sleep(EXPIRE * 1.1)
        assert await func(b"notok") == b"notok"
        assert len(detector.keys) == 1

    assert decorators.context_cache_detect._levels == {}


async def test_context_cache_detect_deep(backend):
    @decorators.cache(backend, ttl=EXPIRE, key="key1")
    async def func1():
        return 1

    @decorators.cache(backend, ttl=EXPIRE, key="key2")
    async def func2():
        return 2

    async def func():
        return await asyncio.gather(func1(), func2())

    with decorators.context_cache_detect as detector:
        await func()
        assert detector.keys == {}

        await asyncio.sleep(0)
        await func()

        assert len(detector.keys) == 2
        assert "key1" in detector.keys
        assert "key2" in detector.keys
    assert decorators.context_cache_detect._levels == {}


async def test_context_cache_detect_context(backend):
    @decorators.cache(backend, ttl=EXPIRE, key="key1")
    async def func1():
        return 1

    @decorators.cache(backend, ttl=EXPIRE, key="key2")
    async def func2():
        return 2

    async def func(*funcs):
        with decorators.context_cache_detect as detector:
            await asyncio.gather(*funcs)
            return len(detector.keys)

    await backend.set("key1", "test")
    await backend.set("key2", "test")
    assert await func1() == "test"
    assert await func2() == "test"

    with decorators.context_cache_detect as detector:
        assert await asyncio.create_task(func(func1())) == 1
        assert await asyncio.create_task(func(func1(), func2())) == 2
        assert await asyncio.create_task(func(func2())) == 1
        assert await asyncio.create_task(func(func2(), func2())) == 1

        assert len(detector.keys) == 2

    assert decorators.context_cache_detect._levels == {}



from datetime import timedelta

import pytest

from cashews import mem

pytestmark = [pytest.mark.integration]


@pytest.fixture(name="app")
def _app():
    from fastapi import FastAPI, Header

    app = FastAPI()

    @app.get("/")
    @mem.early(ttl=timedelta(seconds=1), key="root:{q}")
    async def root(q: str = "q", x_name: str = Header("test")):
        return {"name": x_name, "q": q}

    return app


@pytest.fixture(name="client")
def _client(app):
    from starlette.testclient import TestClient

    return TestClient(app)


def test_no_cache(client):
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"name": "test", "q": "q"}

    response = client.get("/?q=test", headers={"X-Name": "name"})
    assert response.status_code == 200
    assert response.json() == {"name": "name", "q": "test"}


def test_cache(client):
    response = client.get("/")
    assert response.status_code == 200
    assert response.json() == {"name": "test", "q": "q"}

    response = client.get("/", headers={"X-Name": "name"})
    assert response.status_code == 200
    assert response.json() == {"name": "test", "q": "q"}



import random
from contextvars import ContextVar
from typing import Any, Dict

from ..._typing import CacheCondition

_empty = object()


def _not_none_store_condition(result, args, kwargs, key=None) -> bool:
    return result is not None


def _store_all(result, args, kwargs, key=None) -> bool:
    return True


NOT_NONE = "not_none"
_ALL_CONDITIONS = {Any, "all", any, "any", None}
_NOT_NONE_CONDITIONS = {NOT_NONE, "skip_none"}


def _get_cache_condition(condition: CacheCondition):
    if condition in _ALL_CONDITIONS:
        return _store_all
    if condition in _NOT_NONE_CONDITIONS:
        return _not_none_store_condition
    return condition


class CacheDetect:
    __slots__ = ("_value", "_unset_token", "_previous_level")

    def __init__(self, previous_level=0, unset_token=None):
        self._value = {}
        self._unset_token = unset_token
        self._previous_level = previous_level

    def _set(self, key: str, **kwargs: Any) -> None:
        self._value.setdefault(key, []).append(kwargs)

    @property
    def calls(self):
        return dict(self._value)

    keys = calls  # backward compatibility

    def clear(self):
        self._value = {}


_level = ContextVar("level", default=0)


class _ContextCacheDetect:
    def __init__(self):
        self._levels: Dict[int, CacheDetect] = {}

    @property
    def level(self):
        return _level.get()

    def _get_next_level(self):
        level = random.random()
        if level not in self._levels:
            return level
        return self._get_next_level()

    def _start(self) -> CacheDetect:
        previous_level = self.level
        level = self._get_next_level()
        token = _level.set(level)
        self._levels[level] = CacheDetect(previous_level=previous_level, unset_token=token)
        return self._levels[level]

    def _set(self, key: str, **kwargs: Any) -> None:
        level = self.level
        while level:
            var = self._levels.get(level)
            if var is None:
                return
            var._set(key, **kwargs)
            level = var._previous_level

    def _stop(self):
        if self.level in self._levels:
            token = self._levels[self.level]._unset_token
            self._levels[self.level].clear()
            del self._levels[self.level]
            _level.reset(token)

    def __enter__(self) -> CacheDetect:
        return self._start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._stop()


context_cache_detect = _ContextCacheDetect()



from .bloom import bloom, dual_bloom  # noqa
from .cache.defaults import NOT_NONE, CacheDetect, context_cache_detect  # noqa
from .cache.early import early  # noqa
from .cache.fail import failover, fast_condition  # noqa
from .cache.hit import hit  # noqa
from .cache.simple import cache  # noqa
from .cache.soft import soft  # noqa
from .circuit_breaker import CircuitBreakerOpen, circuit_breaker  # noqa
from .locked import locked  # noqa
from .rate import RateLimitException, rate_limit  # noqa



from .key import get_call_values
from .utils import get_obj_size


def add_prefix(prefix: str):
    async def _middleware(call, *args, backend=None, cmd=None, **kwargs):
        if cmd.lower() == "get_many":
            return await call(*[prefix + key for key in args])
        call_values = get_call_values(call, args, kwargs)
        as_key = "key"
        if cmd in ["delete_match", "get_match", "keys_match"]:
            as_key = "pattern"
        key = call_values.get(as_key)
        if key:
            call_values[as_key] = prefix + key
            return await call(**call_values)
        return await call(*args, **kwargs)

    return _middleware


def all_keys_lower():
    async def _middleware(call, *args, backend=None, cmd=None, **kwargs):
        if cmd.lower() == "get_many":
            return await call(*[key.lower() for key in args])
        call_values = get_call_values(call, args, kwargs)
        as_key = "key"
        if cmd == "delete_match":
            as_key = "pattern"
        key = call_values.get(as_key)
        if key:
            call_values[as_key] = key.lower()
            return await call(**call_values)
        return await call(*args, **kwargs)

    return _middleware


def memory_limit(min=0, max=None):
    async def _memory_middleware(call, *args, backend=None, cmd=None, **kwargs):
        if cmd != "set":
            return await call(*args, **kwargs)
        call_values = get_call_values(call, args, kwargs)
        value_size = get_obj_size(call_values["value"])
        if max and value_size > max or value_size < min:
            return None
        return await call(*args, **kwargs)

    return _memory_middleware



import logging
from datetime import datetime, timedelta
from functools import wraps
from typing import Optional, Tuple, Type, Union

from ..._typing import CacheCondition
from ...backends.interface import Backend
from ...formatter import register_template
from ...key import get_cache_key, get_cache_key_template
from .defaults import CacheDetect, _empty, _get_cache_condition, context_cache_detect

__all__ = ("soft",)
logger = logging.getLogger(__name__)


def soft(
    backend: Backend,
    ttl: int,
    key: Optional[str] = None,
    soft_ttl: Optional[int] = None,
    exceptions: Union[Type[Exception], Tuple[Type[Exception]]] = Exception,
    condition: CacheCondition = None,
    prefix: str = "soft",
):
    """
    Cache strategy that allow to use pre-expiration

    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key, default 'early'
    """
    store = _get_cache_condition(condition)
    if soft_ttl is None:
        soft_ttl = ttl * 0.33  # type: ignore[assignment]

    def _decor(func):
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                soft_expire_at, result = cached
                if soft_expire_at > datetime.utcnow():
                    _from_cache._set(
                        _cache_key,
                        ttl=ttl,
                        soft_ttl=soft_ttl,
                        name="soft",
                        template=_key_template,
                    )
                    return result

            try:
                result = await func(*args, **kwargs)
            except exceptions:
                if cached is not _empty:
                    _, result = cached
                    _from_cache._set(
                        _cache_key,
                        ttl=ttl,
                        soft_ttl=soft_ttl,
                        name="soft",
                        template=_key_template,
                    )
                    return result
                raise
            else:
                if store(result, args, kwargs, _cache_key):
                    soft_expire_at = datetime.utcnow() + timedelta(seconds=soft_ttl)
                    await backend.set(_cache_key, [soft_expire_at, result], expire=ttl)
                return result

        return _wrap

    return _decor



import asyncio
import math
import warnings
from functools import wraps
from typing import Any, Optional, Tuple, Union

from cashews.utils import get_hashes

from ..backends.interface import Backend
from ..key import get_cache_key, get_cache_key_template

__all__ = ("bloom",)

all_not_zeros = all
all_zeros = lambda values: all((v == 0 for v in values))
at_least_one_zero = lambda values: not all(values)


def bloom(
    backend: Backend,
    name: Optional[str] = None,
    index_size: Optional[int] = None,
    number_of_hashes: Optional[int] = None,
    false_positives: Optional[Union[float, int]] = 1,
    capacity: Optional[int] = None,
    check_false_positive: bool = True,
    prefix: str = "bloom",
):
    """
    Decorator that can help you to use bloom filter algorithm

    @bloom(name="user_name:{name}", false_positives=1, capacity=10_000)
    async def is_user_exists(name):
        return name in ....

    :param backend: cache backend
    :param name: custom cache key
    :param index_size: size of bloom filter
    :param number_of_hashes: the same as k
    :param capacity: the same as n - number of elements
    :param false_positives: Percents of false positive results
    :param check_false_positive: do we need to check if we have positive result
    :param prefix: custom prefix for key, default 'bloom'
    """

    if backend.name == "mem":
        warnings.warn("Remember that you should fill a bloom filter before using it.")

    if index_size is None:
        assert false_positives and capacity
        assert 0 < false_positives < 100
        index_size, number_of_hashes = params_for(capacity, false_positives / 100)

    def _decor(func):
        _name = get_cache_key_template(func, key=name)
        _cache_key = f"{_name}:{index_size}"
        if prefix:
            _cache_key = f"{prefix}:{_cache_key}"

        _set = getattr(func, "set", None)

        async def __set(*args: Any, **kwargs: Any):
            if _set is None:
                result = await func(*args, **kwargs)
            else:
                result = await _set(*args, **kwargs)
            if not result:
                return result
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            hashes = get_hashes(_bloom_key, number_of_hashes, index_size)
            await backend.incr_bits(_cache_key, *hashes)
            return result

        func.set = __set

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            hashes = get_hashes(_bloom_key, number_of_hashes, index_size)
            values = await backend.get_bits(_cache_key, *hashes)
            if values is None:
                return await func(*args, **kwargs)
            if all_not_zeros(values):  # if all bits is set
                # false positive
                if check_false_positive:
                    return await func(*args, **kwargs)
                return True
            else:
                return False

        return _wrap

    return _decor


def dual_bloom(
    backend: Backend,
    name: Optional[str] = None,
    index_size: Union[int, Tuple[int, int], None] = None,
    number_of_hashes: Union[int, Tuple[int, int], None] = None,
    false: Union[float, Tuple[float, float], None] = 1,
    capacity: Union[int, Tuple[int, int], None] = None,
    no_collisions: bool = False,
    prefix: str = "dual_bloom",
):
    """
    Decorator that can help you to use bloom filter algorithm
     but this implementation with 2 bloom filters - one for true and 1 for false

    @dual_bloom(name="user_name:{name}", false_positives=1, capacity=10_000)
    async def is_user_exists(name):
        return name in ....

    :param backend: cache backend
    :param name: custom cache key
    :param index_size: size of bloom filter
    :param number_of_hashes: the same as k
    :param capacity: the same as n - number of elements
    :param no_collisions: add value only no collisions
    :param false: Percents of false results
    :param prefix: custom prefix for key, default 'dual_bloom'
    """
    filters_params = _get_params_for_filters(index_size, number_of_hashes, false, capacity)

    def _decor(func):
        _cache_key = get_cache_key_template(func, key=name)
        if prefix:
            _cache_key = f"{prefix}:{_cache_key}"
        _true_bloom_key = _cache_key + ":true"
        _false_bloom_key = _cache_key + ":false"

        __delete = getattr(func, "delete", None)

        async def _delete(*args: Any, **kwargs: Any) -> None:
            if __delete:
                await __delete(*args, **kwargs)
            _bloom_key = get_cache_key(func, _cache_key, args, kwargs)
            hashes_true, hashes_false = _get_hashes(_bloom_key, filters_params)
            await asyncio.gather(
                backend.incr_bits(_true_bloom_key, *hashes_true, by=-1),
                backend.incr_bits(_false_bloom_key, *hashes_false, by=-1),
            )

        func.delete = _delete

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _bloom_key = get_cache_key(func, _cache_key, args, kwargs)
            hashes_true, hashes_false = _get_hashes(_bloom_key, filters_params)
            true_values, false_values = await asyncio.gather(
                backend.get_bits(_true_bloom_key, *hashes_true),
                backend.get_bits(_false_bloom_key, *hashes_false),
            )
            if at_least_one_zero(true_values) and at_least_one_zero(false_values):
                # not set yet
                result = await func(*args, **kwargs)
                if result and (not no_collisions or all_zeros(true_values)):
                    await backend.incr_bits(_true_bloom_key, *hashes_true)
                if not result and (not no_collisions or all_zeros(false_values)):
                    await backend.incr_bits(_false_bloom_key, *hashes_false)
                return result
            if at_least_one_zero(true_values) and all(false_values):
                return False  # can be false Negative
            if at_least_one_zero(false_values) and all(true_values):
                return True  # can be false Positive
            return await func(*args, **kwargs)

        return _wrap

    return _decor


def _get_params_for_filters(
    index_size: Union[int, Tuple[int, int], None] = None,
    number_of_hashes: Union[int, Tuple[int, int], None] = None,
    false: Union[float, Tuple[float, float], None] = 1,
    capacity: Union[int, Tuple[int, int], None] = None,
):
    if index_size is None:
        assert false and capacity
        assert 0 < false < 100
        if isinstance(capacity, int):
            capacity = (capacity, capacity)
        if isinstance(false, int):
            false = (false, false)
        index_size_false, _number_of_hashes_false = params_for(capacity[0], false[0] / 100)
        index_size_true, _number_of_hashes_true = params_for(capacity[1], false[1] / 100)
        index_size = (index_size_true, index_size_false)
        number_of_hashes = (_number_of_hashes_true, _number_of_hashes_false)
    else:
        assert number_of_hashes
        if isinstance(index_size, int):
            index_size = (index_size, index_size)
        if isinstance(number_of_hashes, int):
            number_of_hashes = (number_of_hashes, number_of_hashes)
    return list(zip(index_size, number_of_hashes))


def _get_hashes(key, filters_params):
    params_true, params_false = filters_params
    index_size_true, number_of_hashes_true = params_true
    index_size_false, number_of_hashes_false = params_false
    hashes_true = get_hashes(key + "true", number_of_hashes_true, index_size_true)
    hashes_false = get_hashes(key + "false", number_of_hashes_false, index_size_false)
    return hashes_true, hashes_false


def _counting_bloom(
    backend: Backend,
    name: Optional[str] = None,
    index_size: Optional[int] = None,
    number_of_hashes: Optional[int] = None,
    false_positives: Optional[Union[float, int]] = 1,
    capacity: Optional[int] = None,
    check_false_positive: bool = True,
    prefix: str = "counting_bloom",
):
    """
    Decorator that can help you to use bloom count filter algorithm
    https://en.wikipedia.org/wiki/Counting_Bloom_filter

    :param backend: cache backend
    :param name: custom cache key
    :param index_size: size of bloom filter
    :param number_of_hashes: the same as k
    :param capacity: the same as n - number of elements
    :param false_positives: Percents of false positive results
    :param prefix: custom prefix for key, default 'counting_bloom'
    """
    if index_size is None:
        assert false_positives and capacity
        assert 0 < false_positives < 100
        index_size, number_of_hashes = params_for(capacity, false_positives / 100)

    def _decor(func):
        _name = get_cache_key_template(func, key=name)
        _cache_key = prefix + _name + f":{index_size}"

        @wraps(func)
        async def _wrap(*args, **kwargs):
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            hashes = get_hashes(_bloom_key, number_of_hashes, index_size)
            values = await backend.get_bits(_cache_key, *hashes, size=2)
            if at_least_one_zero(values):
                return False
            if all(values):
                if not check_false_positive:
                    return True
            result = await func(*args, **kwargs)
            assert isinstance(result, bool)
            if result:
                await backend.incr_bits(_cache_key, *hashes, size=2, by=3)
            else:
                await backend.incr_bits(_cache_key, *hashes, size=2, by=1)
            return result

        async def _delete(*args: Any, **kwargs: Any) -> None:
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            hashes = get_hashes(_bloom_key, number_of_hashes, index_size)
            await backend.incr_bits(_cache_key, *hashes, size=2, by=-1)

        func.delete = _wrap.delete = _delete

        async def _set(*args: Any, **kwargs: Any) -> None:
            _bloom_key = get_cache_key(func, _name, args, kwargs)
            hashes = get_hashes(_bloom_key, number_of_hashes, index_size)
            await backend.incr_bits(_cache_key, *hashes, size=2, by=1)

        func.set = _wrap.set = _set

        return _wrap

    return _decor


def params_for(size: int, false_positives: float = 0.01):
    """
    size = n
    k=(m * ln 2 / n)
    m = k * n / ln 2
    p = (1-e^{-k * n/m})^{k}
    return m - number of bits in the array and k the number of hash functions
    """
    m = _count_m(size, false_positives)
    k = _count_k(m, size)
    return m, k


def _count_k(m, n):
    return int(round((m / n) * math.log(2)))


def _count_k_from_p(p):
    return int(math.ceil(math.log(1.0 / p, 2)))


def _count_m(n, p=0.1):
    return int(math.ceil(-1 * n * math.log(p) / math.log(2) ** 2))


def _count_probability(n, m, k):
    ev = -(k * n / m)
    ome = 1 - math.e**ev
    return math.pow(ome, k)



import asyncio

import pytest

from cashews.backends.memory import Memory

pytestmark = [pytest.mark.asyncio, pytest.mark.redis]


@pytest.fixture(name="create_cache")
def _create_cache(redis_dsn, backend_factory):
    from cashews.backends.client_side import BcastClientSide

    async def call(local_cache):
        return await backend_factory(BcastClientSide, redis_dsn, hash_key=None, local_cache=local_cache)

    return call


async def test_set_get_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("key", b"value", expire=0.1)
    await asyncio.sleep(0.01)  # skip init signal about invalidation
    assert await cachef.get("key") == b"value"
    assert await caches.get("key") == b"value"
    assert await cachef_local.get("key") == b"value"
    assert await caches_local.get("key") == b"value"
    await asyncio.sleep(0.2)
    assert await cachef_local.get("key") is None
    assert await caches_local.get("key") is None

    assert await caches.get("key") is None


async def test_set_none_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("key", None, expire=0.1)
    await asyncio.sleep(0.01)  # skip init signal about invalidation
    assert await cachef_local.exists("key")
    assert await cachef.exists("key")

    assert await caches.get("key") is None
    assert await caches.exists("key")
    assert await caches_local.exists("key")


async def test_del_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("key", b"value")
    await asyncio.sleep(0.05)  # skip init signal about invalidation

    assert await cachef.get("key") == b"value"
    assert await caches.get("key") == b"value"
    await cachef.delete("key")
    await asyncio.sleep(0.05)  # skip init signal about invalidation
    assert await caches.get("key") is None


async def test_rewrite_bcast(create_cache):
    cachef_local = Memory()
    cachef = await create_cache(cachef_local)
    caches_local = Memory()
    caches = await create_cache(caches_local)

    await cachef.set("key", b"value")
    await asyncio.sleep(0.05)  # skip init signal about invalidation

    assert await cachef.get("key") == b"value"
    assert await caches.get("key") == b"value"

    await caches.set("key", b"new", expire=0.1)
    await asyncio.sleep(0.05)  # skip init signal about invalidation

    assert await cachef.get("key") == b"new"

    await asyncio.sleep(0.15)
    assert await caches.get("key") is None
    assert await cachef.get("key") is None


async def test_simple_cmd_bcast(create_cache):
    local = Memory()
    cache = await create_cache(local)

    await cache.incr("key:2")
    assert await cache.get("key:2") == 1
    await cache.delete("key:2")
    assert await cache.get("key:2") is None
    assert await local.get("key:2") is None

    await cache.set("key:1", "test", 10)
    assert await cache.get("key:1") == "test"
    assert await local.get("key:1") == "test"
    await cache.expire("key:1", 100)
    assert await cache.get_expire("key:1") > 10
    assert await local.get_expire("key:1") > 10
    assert await cache.get("key:1") == "test"
    assert await local.get("key:1") == "test"

    await cache.clear()
    cache.close()


async def test_simple_cmd_bcast_many(create_cache):
    local = Memory()
    cache = await create_cache(local)
    await cache.set("key:1", "test")
    assert await cache.get("key:1") == "test"
    assert await local.get("key:1") == "test"

    assert await cache.get_many("key:1", "key:2") == ("test", None)

    async for key in cache.scan("key:*"):
        assert key == "key:1"
        break
    else:
        assert False

    async for key, value in cache.get_match("key:*"):
        assert key == "key:1"
        assert value == "test"
        break
    else:
        assert False

    await local.clear()

    assert await cache.get_many("key:1", "key:2") == ("test", None)

    async for key in cache.scan("key:*"):
        assert key == "key:1"

    async for key, value in cache.get_match("key:*"):
        assert key == "key:1"
        assert value == "test"

    assert await local.get("key:1") == "test"

    assert await cache.delete_match("key:*")
    assert await cache.get("key:1") is None
    assert await local.get("key:1") is None

    async for _ in cache.scan("key:*"):
        assert False

    async for _ in cache.get_match("key:*"):
        assert False



import gc
import sys


def get_obj_size(obj) -> int:
    marked = {id(obj)}
    obj_q = [obj]
    size = 0

    while obj_q:
        size += sum(map(sys.getsizeof, obj_q))

        # Lookup all the object referred to by the object in obj_q.
        # See: https://docs.python.org/3.7/library/gc.html#gc.get_referents
        all_refr = ((id(o), o) for o in gc.get_referents(*obj_q))

        # Filter object that are already marked.
        # Using dict notation will prevent repeated objects.
        new_refr = {o_id: o for o_id, o in all_refr if o_id not in marked and not isinstance(o, type)}

        # The new obj_q will be the ones that were not marked,
        # and we will update marked with their ids so we will
        # not traverse them again.
        obj_q = new_refr.values()
        marked.update(new_refr.keys())

    return size



import hashlib
import zlib

algorithms = [
    zlib.crc32,
]

try:
    import xxhash
except ImportError:
    pass
else:
    algorithms.extend(
        [xxhash.xxh3_64_intdigest, xxhash.xxh64_intdigest, xxhash.xxh3_128_intdigest, xxhash.xxh32_intdigest]
    )


def get_hashes(key: str, k: int, max_i: int):
    """
    return array with bit indexes for given value (key) [23, 45, 15]
    """

    assert max_i >= k

    indexes = set()
    for i in range(k):
        ii = i % len(algorithms)

        value = algorithms[ii](f"{key}_{i}".encode()) % max_i
        while value in indexes:
            i += 1
            value = algorithms[ii](f"{key}_{i}".encode()) % max_i
        indexes.add(value)
    return indexes
    # str_hash = str(_get_string_int_hash(key))
    # indexes = set()
    # for _hash in _split_string_for_chunks(str_hash, k):
    #     value = int(_hash) % max_i
    #     while value in indexes:
    #         value += 1
    #     indexes.add(value)
    # return indexes


def _get_string_int_hash(key):
    return int(hashlib.sha256(key.encode("utf-8")).hexdigest(), 16)


def _split_string_for_chunks(value: str, chunks: int):
    chunk_size = len(value) // chunks
    return (value[i : i + chunk_size] for i in range(0, chunk_size * chunks, chunk_size))



import hashlib
import hmac
import pickle
from contextlib import suppress
from typing import Any, Optional, Tuple, Union

BLANK_DIGEST = b""


class UnSecureDataError(Exception):
    pass


class SignIsMissingError(Exception):
    ...


class PickleSerializerMixin:
    _digestmods = {
        b"sha1": hashlib.sha1,
        b"md5": hashlib.md5,
        b"sha256": hashlib.sha256,
    }

    def __init__(
        self,
        *args,
        hash_key: Union[str, bytes, None] = None,
        digestmod: Union[str, bytes, None] = b"md5",
        check_repr: bool = True,
        **kwargs: Any,
    ):
        super().__init__(*args, **kwargs)
        if hash_key is None:
            digestmod = BLANK_DIGEST
            self._digestmods = self._digestmods.copy()
            self._digestmods[BLANK_DIGEST] = lambda value: BLANK_DIGEST  # type: ignore[misc, assignment]
        self._hash_key = _to_bytes(hash_key)
        self._digestmod = _to_bytes(digestmod)
        self._check_repr = check_repr

    async def get(self, key: str, default: Any = None):
        return await self._get_value(await super().get(key), key, default=default)

    async def _get_value(self, value, key, default=None) -> Any:
        try:
            return self._process_value(value, key, default=default)
        except (pickle.PickleError, AttributeError):
            return default

    def _process_value(self, value: Union[None, int, bytes], key: str, default=None) -> Any:
        if value is None:
            return default
        if isinstance(value, int):
            return value
        if value.isdigit():
            return int(value)
        try:
            value = self._split_value_from_signature(value, key)
        except SignIsMissingError:
            return value
        value = pickle.loads(value, fix_imports=False, encoding="bytes")
        if self._check_repr:
            repr(value)
        return value

    def _split_value_from_signature(self, value: bytes, key: str) -> bytes:
        if self._hash_key:
            try:
                sign, value = value.split(b"_", 1)
            except ValueError:
                raise SignIsMissingError(f"key: {key}")
            sign, digestmod = self._get_digestmod(sign)
            expected_sign = self._get_sign(key, value, digestmod)
            if expected_sign != sign:
                raise UnSecureDataError(f"{expected_sign!r} != {sign!r}")
            return value
        else:
            # Backward compatibility.
            DeprecationWarning(
                "If a sign is not used to secure your data, then a value will be pickled and saved without an empty sign prepended."
                "Values saved via 4.x package version without using a sign will not be compatible after the 5.x release."
            )
            with suppress(ValueError):
                sign, value = value.split(b"_", 1)
                if sign:
                    value = sign + b"_" + value
            return value

    def _get_digestmod(self, sign: bytes) -> Tuple[bytes, bytes]:
        digestmod = self._digestmod
        if b":" in sign:
            digestmod, sign = sign.split(b":")
        if digestmod not in self._digestmods:
            raise UnSecureDataError()
        return sign, digestmod

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Any:
        values = []
        for key, value in zip(keys, await super().get_many(*keys, default=default) or [None] * len(keys)):
            values.append(await self._get_value(value, key))
        return tuple(values)

    async def set(self, key: str, value: Any, *args: Any, **kwargs: Any):
        if isinstance(value, int) and not isinstance(value, bool):
            return await super().set(key, value, *args, **kwargs)
        value = pickle.dumps(value, protocol=pickle.HIGHEST_PROTOCOL, fix_imports=False)
        return await super().set(key, self._prepend_sign_to_value(key, value), *args, **kwargs)

    def _prepend_sign_to_value(self, key: str, value: bytes) -> bytes:
        sign = self._get_sign(key, value, self._digestmod)
        if not sign:
            return value
        return self._digestmod + b":" + sign + b"_" + value

    def _get_sign(self, key: str, value: bytes, digestmod: bytes) -> bytes:
        if digestmod == BLANK_DIGEST:
            return BLANK_DIGEST
        value = key.encode() + value
        return hmac.new(self._hash_key, value, self._digestmods[digestmod]).hexdigest().encode()

    def set_raw(self, *args: Any, **kwargs: Any):
        return super().set(*args, **kwargs)

    def get_raw(self, *args: Any, **kwargs: Any):
        return super().get(*args, **kwargs)


def _to_bytes(value: Union[str, bytes, None]) -> Optional[bytes]:
    if value is None:
        return None
    if isinstance(value, str):
        value = value.encode()
    return value



import pytest

from cashews.backends.memory import Memory
from cashews.wrapper import BackendNotAvailable, settings_url_parse


@pytest.mark.parametrize(
    ("url", "params"),
    (
        ("://", {"backend": Memory, "disable": True}),
        ("mem://", {"backend": Memory}),
        (
            "mem://?size=10&check_interval=0.01",
            {"backend": Memory, "size": 10, "check_interval": 0.01},
        ),
    ),
)
def test_url(url, params):
    assert settings_url_parse(url) == params


@pytest.mark.parametrize(
    ("url", "error"),
    (
        (
            "redis://localhost:9000/0",
            "Redis backend requires `redis` (or `aioredis`) to be installed.",
        ),
        ("disk://", "Disk backend requires `diskcache` to be installed."),
    ),
)
def test_url_but_backend_dependency_is_not_installed(url, error):
    with pytest.raises(BackendNotAvailable) as excinfo:
        settings_url_parse(url)

    assert str(excinfo.value) == error


@pytest.mark.redis
@pytest.mark.parametrize(
    ("url", "params"),
    (
        (
            "redis://localhost:9000/0",
            {"backend": None, "address": "redis://localhost:9000/0"},
        ),
        (
            "redis://password@localhost:9000",
            {"backend": None, "address": "redis://password@localhost:9000"},
        ),
        (
            "redis://localhost/0/?password=password",
            {
                "backend": None,
                "address": "redis://localhost/0/",
                "password": "password",
            },
        ),
        (
            "redis://localhost/0/?hash_key=secret&password=test&safe=1&minsize=3&create_connection_timeout=0.1",
            {
                "backend": None,
                "address": "redis://localhost/0/",
                "hash_key": "secret",
                "password": "test",
                "safe": True,
                "minsize": 3,
                "create_connection_timeout": 0.1,
            },
        ),
        (
            "redis://localhost:9000?",
            {"backend": None, "address": "redis://localhost:9000"},
        ),
    ),
)
def test_url_with_redis_as_backend(url, params):
    from cashews.backends.redis import Redis

    params["backend"] = Redis
    assert settings_url_parse(url) == params


@pytest.mark.diskcache
@pytest.mark.parametrize(
    ("url", "params"),
    (
        ("disk://", {"backend": None}),
        ("disk://?size_limit=1000", {"backend": None, "size_limit": 1000}),
        (
            "disk://?directory=/tmp/cache&timeout=1&shards=0",
            {"backend": None, "directory": "/tmp/cache", "timeout": 1, "shards": 0},
        ),
    ),
)
def test_url_with_diskcache_as_backend(url, params):
    from cashews.backends.diskcache import DiskCache

    params["backend"] = DiskCache
    assert settings_url_parse(url) == params



import asyncio
from contextlib import contextmanager
from functools import partial, wraps
from typing import Any, AsyncIterator, Callable, Dict, Iterable, Optional, Tuple, Type, Union
from urllib.parse import parse_qsl, urlparse

from . import decorators, validation
from ._typing import TTL, CacheCondition
from .backends.interface import Backend
from .backends.memory import Memory
from .disable_control import ControlMixin, _is_disable_middleware
from .key import ttl_to_seconds

try:
    try:
        from redis import asyncio as aioredis
    except ImportError:
        import aioredis
except ImportError:
    BcastClientSide, Redis = None, None
else:
    from .backends.client_side import BcastClientSide
    from .backends.redis import Redis

    del aioredis

try:
    import diskcache
except ImportError:
    DiskCache = None
else:
    from .backends.diskcache import DiskCache


#  pylint: disable=too-many-public-methods


class BackendNotAvailable(Exception):
    pass


def _create_auto_init():
    lock = asyncio.Lock()

    async def _auto_init(call, *args, backend=None, cmd=None, **kwargs):
        if backend.is_init:
            return await call(*args, **kwargs)
        async with lock:
            if not backend.is_init:
                await backend.init()

        return await call(*args, **kwargs)

    return _auto_init


class Cache(Backend):
    default_prefix = ""

    def __init__(self, name: Optional[str] = None):
        self._backends: Dict[str, Tuple[Backend, Tuple[Callable]]] = {}  # {key: (backend, middleware)}
        self._default_middlewares = (
            _is_disable_middleware,
            _create_auto_init(),
            validation._invalidate_middleware,
        )
        self._name = name
        self._default_fail_exceptions = Exception
        self._add_backend(Memory)

    detect = decorators.context_cache_detect

    def set_default_fail_exceptions(self, *exc: Type[Exception]):
        self._default_fail_exceptions = exc

    def disable(self, *cmds: str, prefix: str = ""):
        return self._get_backend(prefix).disable(*cmds)

    def disable_all(self, *cmds: str):
        for backend, _ in self._backends.values():
            backend.disable(*cmds)

    def enable(self, *cmds: str, prefix: str = ""):
        return self._get_backend(prefix).enable(*cmds)

    def enable_all(self, *cmds: str):
        for backend, _ in self._backends.values():
            backend.enable(*cmds)

    @contextmanager
    def disabling(self, *cmds: str, prefix: str = ""):
        self.disable(*cmds, prefix=prefix)
        yield
        self.enable(*cmds, prefix=prefix)

    def is_disable(self, *cmds: str, prefix: str = ""):
        return self._get_backend(prefix).is_disable(*cmds)

    def is_enable(self, *cmds: str, prefix: str = ""):
        return not self.is_disable(*cmds, prefix=prefix)

    def _get_backend_and_config(self, key: str) -> Tuple[Backend, Tuple[Callable]]:
        for prefix in sorted(self._backends.keys(), reverse=True):
            if key.startswith(prefix):
                return self._backends[prefix]
        return self._backends[self.default_prefix]

    def _get_backend(self, key: str) -> Backend:
        backend, _ = self._get_backend_and_config(key)
        return backend

    def setup(self, settings_url: str, middlewares: Tuple = (), prefix: str = default_prefix, **kwargs):
        params = settings_url_parse(settings_url)
        params.update(kwargs)

        if params.pop("client_side", None):
            params["backend"] = BcastClientSide
        backend = params.pop("backend")

        self._add_backend(backend, middlewares, prefix, **params)
        return self._backends[prefix][0]

    def _add_backend(self, backend_class: Type[Backend], middlewares=(), prefix: str = default_prefix, **params):
        class _backend_class(ControlMixin, backend_class):
            pass

        self._backends[prefix] = (
            _backend_class(**params),
            self._default_middlewares + middlewares,
        )

    async def init(self, *args, **kwargs):
        if args or kwargs:
            self.setup(*args, **kwargs)
        for backend, _ in self._backends.values():
            await backend.init()

    def _with_middlewares(self, cmd: str, key):
        backend, middlewares = self._get_backend_and_config(key)
        return self._with_middlewares_for_backend(cmd, backend, middlewares)

    def _with_middlewares_for_backend(self, cmd: str, backend, middlewares):
        call = getattr(backend, cmd)
        for middleware in middlewares:
            call = partial(middleware, call, cmd=cmd, backend=backend)
        return call

    def set(
        self,
        key: str,
        value: Any,
        expire: Union[float, TTL, None] = None,
        exist: Optional[bool] = None,
    ) -> bool:
        return self._with_middlewares("set", key)(key=key, value=value, expire=ttl_to_seconds(expire), exist=exist)

    def set_raw(self, key: str, value: Any, **kwargs):
        return self._with_middlewares("set_raw", key)(key=key, value=value, **kwargs)

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        return self._with_middlewares("get", key)(key=key, default=default)

    def get_raw(self, key: str) -> Any:
        return self._with_middlewares("get_raw", key)(key=key)

    async def keys_match(self, pattern: str):
        backend, middlewares = self._get_backend_and_config(pattern)

        async def call(_pattern):
            return backend.keys_match(_pattern)

        for middleware in middlewares:
            call = partial(middleware, call, cmd="keys_match", backend=backend)
        async for key in (await call(pattern)):
            yield key

    async def scan(self, pattern: str, batch_size: int = 100) -> AsyncIterator[str]:
        backend, middlewares = self._get_backend_and_config(pattern)

        async def call(_pattern):
            return backend.scan(_pattern)

        for middleware in middlewares:
            call = partial(middleware, call, cmd="scan", backend=backend)
        async for key in (await call(pattern)):
            yield key

    async def get_match(
        self, pattern: str, batch_size: int = 100, default: Optional[Any] = None
    ) -> AsyncIterator[Tuple[str, Any]]:
        backend, middlewares = self._get_backend_and_config(pattern)

        async def call(_pattern, _batch_size, _default):
            return backend.get_match(_pattern, batch_size=_batch_size, default=_default)

        for middleware in middlewares:
            call = partial(middleware, call, cmd="get_match", backend=backend)
        async for key, value in (await call(pattern, batch_size, default)):
            yield key, value

    async def get_many(self, *keys: str, default: Optional[Any] = None) -> Tuple[Any]:
        backends = {}
        for key in keys:
            backend = self._get_backend(key)
            backends.setdefault(backend, []).append(key)
        result = {}
        for _keys in backends.values():
            _values = await self._with_middlewares("get_many", _keys[0])(*_keys, default=default)
            result.update(dict(zip(_keys, _values)))
        return tuple([result.get(key) for key in keys])

    def get_bits(self, key: str, *indexes: int, size: int = 1) -> Tuple[int]:
        return self._with_middlewares("get_bits", key)(key, *indexes, size=size)

    def incr_bits(self, key: str, *indexes: int, size: int = 1, by: int = 1):
        return self._with_middlewares("incr_bits", key)(key, *indexes, size=size, by=by)

    def incr(self, key: str) -> int:
        return self._with_middlewares("incr", key)(key=key)

    def delete(self, key: str):
        return self._with_middlewares("delete", key)(key=key)

    def delete_match(self, pattern: str):
        return self._with_middlewares("delete_match", pattern)(pattern=pattern)

    def expire(self, key: str, timeout: TTL):
        return self._with_middlewares("expire", key)(key=key, timeout=ttl_to_seconds(timeout))

    def get_expire(self, key: str):
        return self._with_middlewares("get_expire", key)(key=key)

    def exists(self, key: str):
        return self._with_middlewares("exists", key)(key=key)

    def set_lock(self, key: str, value: Any, expire: TTL):
        return self._with_middlewares("set_lock", key)(key=key, value=value, expire=ttl_to_seconds(expire))

    def unlock(self, key: str, value: str):
        return self._with_middlewares("unlock", key)(key=key, value=value)

    def get_size(self, key: str):
        return self._with_middlewares("get_size", key)(key)

    def ping(self, message: Optional[bytes] = None) -> str:
        message = b"PING" if message is None else message
        return self._with_middlewares("ping", message.decode())(message=message)

    async def clear(self):
        for backend, _ in self._backends.values():
            await self._with_middlewares_for_backend("clear", backend, self._default_middlewares)()

    def close(self):
        for backend, _ in self._backends.values():
            backend.close()

    def is_locked(
        self,
        key: str,
        wait: Union[float, None, TTL] = None,
        step: Union[int, float] = 0.1,
    ) -> bool:
        return self._with_middlewares("is_locked", key)(key=key, wait=ttl_to_seconds(wait), step=step)

    def _wrap_on(self, decorator_fabric, upper, **decor_kwargs):
        if upper:
            return self._wrap_with_condition(decorator_fabric, **decor_kwargs)
        return self._wrap(decorator_fabric, **decor_kwargs)

    def _wrap(self, decorator_fabric, lock=False, **decor_kwargs):
        def _decorator(func):
            decorator = decorator_fabric(self, **decor_kwargs)(func)

            @wraps(func)
            async def _call(*args, **kwargs):
                if lock:
                    _locked = decorators.locked(self, key=decor_kwargs.get("key"), ttl=decor_kwargs["ttl"])
                    return await _locked(decorator)(*args, **kwargs)
                else:
                    return await decorator(*args, **kwargs)

            _call.direct = func
            return _call

        return _decorator

    def _wrap_with_condition(self, decorator_fabric, condition, lock=False, **decor_kwargs):
        def _decorator(func):
            decorator_fabric(self, **decor_kwargs)(func)  # to register cache templates

            @wraps(func)
            async def _call(*args, **kwargs):
                with decorators.context_cache_detect as detect:

                    def new_condition(result, _args, _kwargs, key):
                        if detect.keys:
                            return False
                        return condition(result, _args, _kwargs, key=key) if condition else result is not None

                    decorator = decorator_fabric(self, **decor_kwargs, condition=new_condition)
                    if lock:
                        _locked = decorators.locked(self, key=decor_kwargs.get("key"), ttl=decor_kwargs["ttl"])
                        result = await _locked(decorator(func))(*args, **kwargs)
                    else:
                        result = await decorator(func)(*args, **kwargs)

                return result

            _call.direct = func
            return _call

        return _decorator

    # DecoratorS
    def __call__(
        self,
        ttl: TTL,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        prefix: str = "",
        upper: bool = False,
        lock: bool = False,
    ):
        return self._wrap_on(
            decorators.cache,
            upper,
            lock=lock,
            ttl=ttl_to_seconds(ttl),
            key=key,
            condition=condition,
            prefix=prefix,
        )

    cache = __call__

    def failover(
        self,
        ttl: TTL,
        exceptions: Union[Type[Exception], Iterable[Type[Exception]], None] = None,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        prefix: str = "fail",
    ):
        exceptions = exceptions or self._default_fail_exceptions
        return self._wrap_with_condition(
            decorators.failover,
            ttl=ttl_to_seconds(ttl),
            exceptions=exceptions,
            key=key,
            condition=condition,
            prefix=prefix,
        )

    def early(
        self,
        ttl: TTL,
        key: Optional[str] = None,
        early_ttl: Optional[TTL] = None,
        condition: CacheCondition = None,
        prefix: str = "early",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.early,
            upper,
            ttl=ttl_to_seconds(ttl),
            key=key,
            early_ttl=ttl_to_seconds(early_ttl),
            condition=condition,
            prefix=prefix,
        )

    def soft(
        self,
        ttl: TTL,
        key: Optional[str] = None,
        soft_ttl: Optional[TTL] = None,
        exceptions: Union[Type[Exception], Tuple[Type[Exception]]] = Exception,
        condition: CacheCondition = None,
        prefix: str = "soft",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.soft,
            upper,
            ttl=ttl_to_seconds(ttl),
            key=key,
            soft_ttl=ttl_to_seconds(soft_ttl),
            exceptions=exceptions,
            condition=condition,
            prefix=prefix,
        )

    def hit(
        self,
        ttl: TTL,
        cache_hits: int,
        update_after: int = 0,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        prefix: str = "hit",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.hit,
            upper,
            ttl=ttl_to_seconds(ttl),
            cache_hits=cache_hits,
            update_after=update_after,
            key=key,
            condition=condition,
            prefix=prefix,
        )

    def dynamic(
        self,
        ttl: TTL = 60 * 60 * 24,
        key: Optional[str] = None,
        condition: CacheCondition = None,
        prefix: str = "dynamic",
        upper: bool = False,
    ):
        return self._wrap_on(
            decorators.hit,
            upper,
            ttl=ttl_to_seconds(ttl),
            cache_hits=3,
            update_after=1,
            key=key,
            condition=condition,
            prefix=prefix,
        )

    def invalidate(
        self,
        func,
        args_map: Optional[Dict[str, str]] = None,
        defaults: Optional[Dict] = None,
    ):
        return validation.invalidate(
            backend=self,
            target=func,
            args_map=args_map,
            defaults=defaults,
        )

    invalidate_func = validation.invalidate_func

    def circuit_breaker(
        self,
        errors_rate: int,
        period: TTL,
        ttl: TTL,
        exceptions: Union[Type[Exception], Tuple[Type[Exception]], None] = None,
        key: Optional[str] = None,
        prefix: str = "circuit_breaker",
    ):
        exceptions = exceptions or self._default_fail_exceptions
        return decorators.circuit_breaker(
            backend=self,
            errors_rate=errors_rate,
            period=ttl_to_seconds(period),
            ttl=ttl_to_seconds(ttl),
            exceptions=exceptions,
            key=key,
            prefix=prefix,
        )

    def rate_limit(
        self,
        limit: int,
        period: TTL,
        ttl: Optional[TTL] = None,
        action: Optional[Callable] = None,
        prefix="rate_limit",
    ):  # pylint: disable=too-many-arguments
        return decorators.rate_limit(
            backend=self,
            limit=limit,
            period=ttl_to_seconds(period),
            ttl=ttl_to_seconds(ttl),
            action=action,
            prefix=prefix,
        )

    def locked(
        self,
        ttl: Optional[TTL] = None,
        key: Optional[str] = None,
        step: Union[int, float] = 0.1,
        prefix: str = "locked",
    ):
        return decorators.locked(
            backend=self,
            ttl=ttl_to_seconds(ttl),
            key=key,
            step=step,
            prefix=prefix,
        )

    def bloom(
        self,
        name: Optional[str] = None,
        index_size: Optional[int] = None,
        number_of_hashes: Optional[int] = None,
        false_positives: Optional[Union[float, int]] = 1,
        capacity: Optional[int] = None,
        check_false_positive: bool = True,
        prefix: str = "bloom",
    ):
        return decorators.bloom(
            backend=self,
            name=name,
            index_size=index_size,
            number_of_hashes=number_of_hashes,
            false_positives=false_positives,
            capacity=capacity,
            check_false_positive=check_false_positive,
            prefix=prefix,
        )

    def dual_bloom(
        self,
        name: Optional[str] = None,
        index_size: Optional[int] = None,
        number_of_hashes: Optional[int] = None,
        false: Optional[Union[float, int]] = 1,
        no_collisions: bool = False,
        capacity: Optional[int] = None,
        prefix: str = "dual_bloom",
    ):
        return decorators.dual_bloom(
            backend=self,
            name=name,
            index_size=index_size,
            number_of_hashes=number_of_hashes,
            false=false,
            no_collisions=no_collisions,
            capacity=capacity,
            prefix=prefix,
        )


def _fix_params_types(params: Dict[str, str]) -> Dict[str, Union[str, int, bool, float]]:
    new_params = {}
    bool_keys = ("safe", "enable", "disable", "client_side")
    true_values = (
        "1",
        "true",
    )
    for key, value in params.items():
        if key.lower() in bool_keys:
            value = value.lower() in true_values
        elif value.isdigit():
            value = int(value)
        else:
            try:
                value = float(value)
            except ValueError:
                pass
        new_params[key.lower()] = value
    return new_params


def settings_url_parse(url):
    params = {}
    parse_result = urlparse(url)
    params.update(dict(parse_qsl(parse_result.query)))
    params = _fix_params_types(params)
    if parse_result.scheme == "redis" or parse_result.scheme == "rediss":
        if Redis is None:
            raise BackendNotAvailable("Redis backend requires `redis` (or `aioredis`) to be installed.")
        params["backend"] = Redis
        params["address"] = parse_result._replace(query=None)._replace(fragment=None).geturl()
    elif parse_result.scheme == "mem":
        params["backend"] = Memory
    elif parse_result.scheme == "disk":
        if DiskCache is None:
            raise BackendNotAvailable("Disk backend requires `diskcache` to be installed.")
        params["backend"] = DiskCache
    elif parse_result.scheme == "":
        params["backend"] = Memory
        params["disable"] = True
    else:
        raise BackendNotAvailable(f"wrong backend alias {parse_result.scheme}")
    return params



import asyncio
from unittest.mock import Mock, PropertyMock

import pytest

from cashews.backends.memory import Memory
from cashews.disable_control import ControlMixin
from cashews.formatter import get_templates_for_func
from cashews.helpers import add_prefix, all_keys_lower
from cashews.wrapper import Cache, _create_auto_init

pytestmark = pytest.mark.asyncio


@pytest.fixture(name="target")
def _target():
    class New(ControlMixin, Memory):
        pass

    return Mock(wraps=New())


@pytest.fixture(name="cache")
def __cache(target):
    _cache = Cache()
    _cache._add_backend(Memory)
    _cache._backends[""] = (target, _cache._backends[""][1])
    return _cache


async def test_init_disable(cache):
    await cache.init("mem://localhost?disable=1")
    assert cache.is_disable()


async def test_prefix(cache):
    await cache.init("mem://localhost")
    await cache.init("://", prefix="-")
    assert not cache.is_disable()
    assert cache.is_disable(prefix="-")

    await cache.set("key", "value")
    await cache.set("-:key", "-value")

    assert await cache.get("key") == "value"
    assert await cache.get("-:key") == None
    assert await cache.get("-:key", default="def") == "def"


async def test_prefix_many(cache):
    await cache.init("mem://")
    await cache.init("mem://", prefix="-")

    await cache.set("key", "value")
    await cache.set("-:key", "-value")

    assert await cache.get("key") == "value"
    assert await cache.get("-:key") == "-value"

    assert await cache.get_many("key", "-:key") == ("value", "-value")


async def test_disable_cmd(cache):
    await cache.init("mem://localhost")
    cache.disable("incr")
    await cache.set("test", 10)
    await cache.incr("test")
    assert await cache.get("test") == 10

    cache.enable("incr")
    await cache.incr("test")
    assert await cache.get("test") == 11


async def test_disable_ctz(cache):
    await cache.init("mem://localhost")
    cache.enable()

    async def test():
        await cache.set("test", "1")
        assert await cache.get("test") == "1"
        cache.disable("set")
        await cache.set("test", "2")

    await asyncio.create_task(test())
    assert await cache.get("test") == "1"
    await cache.set("test", "3")
    assert await cache.get("test") == "3"


async def test_disable_decorators(cache: Cache, target):
    cache.disable()
    data = (i for i in range(10))

    @cache(ttl=1)
    @cache.soft(ttl=1)
    @cache.failover(ttl=1)
    @cache.hit(ttl=1, cache_hits=1)
    @cache.circuit_breaker(ttl=1, errors_rate=1, period=1)
    @cache.rate_limit(ttl=1, limit=1, period=1)
    @cache.early(ttl=1)
    @cache.dynamic()
    @cache.locked(ttl=1)
    async def func():
        return next(data)

    assert await func() == 0
    assert await func() == 1
    target.get.assert_not_called()
    target.set.assert_not_called()

    cache.enable()
    assert await func() == 2
    assert await func() == 2


async def test_disable_bloom(cache: Cache, target):
    cache.disable()

    @cache.bloom(index_size=10, number_of_hashes=1)
    async def func():
        return True

    await func.set()
    assert await func()
    target.incr_bits.assert_not_called()
    target.get_bits.assert_not_called()

    cache.enable()
    await func.set()
    assert await func()
    target.incr_bits.assert_called()
    target.get_bits.assert_called()


async def test_disable_decorators_get(cache: Cache):
    data = (i for i in range(10))
    await cache.init("mem://localhost")

    @cache(ttl=1)
    async def func():
        return next(data)

    assert cache.is_enable()
    assert cache.is_enable("set", prefix="cache")
    assert cache.is_enable("get", prefix="cache")
    assert cache.is_enable("set", prefix="")
    assert await func() == 0
    assert await func() == 0

    cache.disable("get")

    assert not cache.is_enable("get")
    assert cache.is_enable("set")

    assert await func() == 1
    assert await func() == 2

    cache.enable("get")
    assert await func() == 2


async def test_disable_decorators_set(cache: Cache):
    data = (i for i in range(10))
    cache.disable("set")

    @cache(ttl=1)
    async def func():
        return next(data)

    assert await func() == 0
    assert await func() == 1

    cache.enable("set")
    assert await func() == 2
    assert await func() == 2


async def test_init(cache):
    await cache.init("mem://localhost")
    assert cache.is_enable()


async def test_auto_init(cache):
    target = Mock(wraps=Memory())
    init = False

    def set_init():
        async def _set():
            nonlocal init
            await asyncio.sleep(0.01)
            init = True

        return _set()

    type(target).is_init = PropertyMock(side_effect=lambda: init)
    target.init.side_effect = set_init
    cache._backends[""] = (target, (_create_auto_init(),))
    await asyncio.gather(cache.ping(), cache.ping(), cache.get("test"))
    target.init.assert_called_once()


async def test_all_keys_lower(cache: Cache, target):
    cache._backends[""] = cache._backends[""][0], (all_keys_lower(),)
    await cache.get(key="KEY")
    target.get.assert_called_once_with(key="key", default=None)

    await cache.set(key="KEY", value="value")
    target.set.assert_called_once_with(
        key="key",
        value="value",
        exist=None,
        expire=None,
    )
    await cache.ping()
    target.ping.assert_called_once_with(message=b"PING")


async def test_add_prefix(cache: Cache, target):
    cache._backends[""] = cache._backends[""][0], (add_prefix("prefix!"),)

    await cache.get(key="key")
    target.get.assert_called_once_with(key="prefix!key", default=None)

    await cache.set(key="key", value="value")
    target.set.assert_called_once_with(
        key="prefix!key",
        value="value",
        exist=None,
        expire=None,
    )
    await cache.ping()
    target.ping.assert_called_once_with(message=b"PING")


async def test_add_prefix_get_many(cache: Cache, target):
    cache._backends[""] = cache._backends[""][0], (add_prefix("prefix!"),)
    await cache.get_many("key")
    target.get_many.assert_called_once_with("prefix!key")


async def test_add_prefix_delete_match(cache: Cache, target):
    cache._backends[""] = cache._backends[""][0], (add_prefix("prefix!"),)
    await cache.delete_match("key")
    target.delete_match.assert_called_once_with(pattern="prefix!key")


async def test_smoke_cmds(cache: Cache, target):
    await cache.set(key="key", value={"any": True}, expire=60, exist=None)
    target.set.assert_called_once_with(key="key", value={"any": True}, expire=60, exist=None)

    await cache.get("key")  # -> Any
    target.get.assert_called_once_with(key="key", default=None)

    await cache.get_many("key1", "key2")
    target.get_many.assert_called_once_with("key1", "key2", default=None)

    await cache.incr("key_incr")  # -> int
    target.incr.assert_called_once_with(key="key_incr")

    await cache.delete("key")
    target.delete.assert_called_once_with(key="key")

    await cache.expire(key="key", timeout=10)
    target.expire.assert_called_once_with(key="key", timeout=10)

    await cache.get_expire(key="key")  # -> int seconds to expire
    target.get_expire.assert_called_once()

    await cache.ping(message=b"test")  # -> bytes
    target.ping.assert_called_once_with(message=b"test")

    await cache.clear()
    target.clear.assert_called_once_with()

    await cache.is_locked("key", wait=60)  # -> bool
    target.is_locked.assert_called_once_with(key="key", wait=60, step=0.1)

    await cache.set_lock("key", "value", expire=60)  # -> bool
    target.set_lock.assert_called_once_with(key="key", value="value", expire=60)

    await cache.unlock("key", "value")  # -> bool
    target.unlock.assert_called_once_with(key="key", value="value")

    await cache.exists("key")
    target.exists.assert_called_once_with(key="key")

    await cache.get_bits("key", 1, 2, 3, size=2)
    target.get_bits.assert_called_once_with("key", 1, 2, 3, size=2)

    await cache.incr_bits("key", 1, 2, 3, size=2)
    target.incr_bits.assert_called_once_with("key", 1, 2, 3, size=2, by=1)

    [key async for key in cache.keys_match("key:*")]
    target.keys_match.assert_called_once_with("key:*")

    [key_value async for key_value in cache.get_match("key:*")]
    target.get_match.assert_called_once_with("key:*", batch_size=100, default=None)


async def test_disable_cache_on_fail_return(cache: Cache):
    assert await cache.get("key") is None

    @cache(ttl=1, key="key", upper=True)
    @cache.failover(ttl=1, key="fail", prefix="")
    async def func(fail=False):
        if fail:
            raise Exception()
        return "1"

    await cache.set("fail", "test_fail")

    assert await func(fail=True) == "test_fail"  # return from fail cache but simple cache should be skipped
    assert await cache.get("key") is None

    await cache.delete("fail")
    await cache.set("key", "val")
    assert await func() == "val"


async def test_disable_cache_on_fail_return_2(cache: Cache):
    assert await cache.get("key") is None

    @cache.failover(ttl=1, key="fail", prefix="")
    @cache(ttl=1, key="key")
    async def func(fail=False):
        if fail:
            raise Exception()
        return "1"

    await cache.set("fail", "test_fail")

    assert await func(fail=True) == "test_fail"  # return from fail cache but simple cache should be skipped
    assert await cache.get("key") is None

    await cache.delete("fail")
    assert await cache.get("fail") is None
    await cache.set("key", "val")
    assert await func() == "val"
    assert await cache.get("fail") is None


async def test_multilayer_cache(cache: Cache):
    # If results from key2, key1 must not be set

    @cache(ttl=1, key="key1", upper=True)
    @cache(ttl=1, key="key2")
    async def func():
        return 1

    await cache.set("key2", "test2")

    assert await func() == "test2"
    assert await cache.get("key1") is None

    await cache.set("key1", "test1")
    assert await func() == "test1"
    assert await cache.get("key2") == "test2"


async def test_cache_decor_register(cache: Cache):
    @cache(ttl=1, key="key:{val}", prefix="test")
    async def my_func(val=1):
        return val

    assert next(get_templates_for_func(my_func)) == "test:key:{val}"


async def test_cache_lock():
    m = Mock()
    cache = Cache()
    cache.setup("mem://")

    @cache(ttl=5, lock=True)
    async def my_func(val=1):
        await asyncio.sleep(0)  # for task switching
        m(val)
        return val

    await asyncio.gather(my_func(), my_func(), my_func())

    m.assert_called_once_with(1)



from datetime import timedelta

import pytest

from cashews.formatter import default_formatter, get_template_and_func_for, get_template_for_key, register_template
from cashews.key import get_cache_key, get_cache_key_template, ttl_to_seconds


async def func1(a):
    ...


async def func2(a, *, k=None, **kwargs):
    ...


async def func3(a, *, k="test"):
    ...


class Klass:
    data = {"test": 1}

    def method(self, a, k=None):
        ...


TEPLATE_FUNC1 = "func1:{a}"
TEPLATE_FUNC2 = "func2:{k}:user:{user}"
TEPLATE_FUNC3 = "func3:{k:len}"


@default_formatter.type_format(Klass)
def _some(value: Klass):
    return "hoho"


@pytest.mark.parametrize(
    ("key", "template", "func"),
    (
        ("func1:test", TEPLATE_FUNC1.format(a="*"), func1),
        ("func1:", TEPLATE_FUNC1.format(a="*"), func1),
        ("prefix:func1:test", TEPLATE_FUNC1.format(a="*"), func1),
        ("func2:-:user:1", TEPLATE_FUNC2.format(k="*", user="*"), func2),
        ("func3:2", "func3:*", func3),
        ("func:1", None, None),
        ("prefix:func2:test:user:1:1", None, None),
        ("func2:user:1", None, None),
        ("func2:user:1", None, None),
    ),
)
def test_detect_template_and_func_by_key(key, template, func):
    register_template(func1, TEPLATE_FUNC1)
    register_template(func2, TEPLATE_FUNC2)
    register_template(func3, TEPLATE_FUNC3)
    _template, _func = get_template_and_func_for(key)
    assert _template == template
    assert _func == ((func.__module__ or "", func.__name__) if func else None)


@pytest.mark.parametrize(
    ("key", "template", "params"),
    (
        ("func1:test", TEPLATE_FUNC1, {"a": "test"}),
        ("func1:", TEPLATE_FUNC1, {"a": ""}),
        ("prefix:func1:test", TEPLATE_FUNC1, {"a": "test"}),
        ("func2:-:user:1", TEPLATE_FUNC2, {"k": "-", "user": "1"}),
        ("func3:2", TEPLATE_FUNC3, {"k": "2"}),
        ("func:1", None, None),
        ("prefix:func2:test:user:1:1", None, None),
        ("func2:user:1", None, None),
        ("func2:user:1", None, None),
    ),
)
def test_detect_template_by_key(key, template, params):
    register_template(func1, TEPLATE_FUNC1)
    register_template(func2, TEPLATE_FUNC2)
    register_template(func3, TEPLATE_FUNC3)

    _template, _params = get_template_for_key(key)
    assert _template == template
    assert _params == params


def test_cache_func_key_dict():
    async def func(user):
        ...

    obj = type("user", (), {"name": "test", "one": True})()

    assert get_cache_key(func, template="test_key:{user.name}", args=(obj,)) == "test_key:test"

    obj = type("user", (), {"name": "new", "one": True})()
    assert get_cache_key(func, template="test_key:{user.name}", args=(obj,)) == "test_key:new"


@pytest.mark.parametrize(
    ("args", "kwargs", "template", "key"),
    (
        (
            ("A1", "a2", "a3"),
            {"kwarg1": "k1", "kwarg3": True},
            "{arg1}:{kwarg1}-{kwarg3}",
            "A1:k1-true",
        ),
        (
            ("A1", "a2", "a3"),
            {"kwarg1": "k1", "kwarg0": True},
            None,
            "tests.test_key:func:arg1:A1:arg2:a2:a3:kwarg1:k1:kwarg2:true:kwarg0:true",
        ),
        (
            ("A1", "a2", "b", True, 1, 1.2, ("a", 1, 1.2)),
            {"kwarg1": "k1", "e": True, "d": 1},
            "{__args__}:{__kwargs__}",
            "b:true:1:1.2:a:1:1.2:d:1:e:true",
        ),
        (
            (b"a1", "a2", "a3"),
            None,
            "{arg1}-{kwarg1}-{kwarg3}",
            "a1--",
        ),
        (
            (True, "a2", "a3"),
            None,
            "{arg1}-{kwarg1}-{kwarg3}",
            "true--",
        ),
        (
            (b"\x16F\xbd\xb0\xcf\xcdN\xd7Y)\xfa\x1d\x96\xb1u\x81", ""),
            None,
            "{arg1}",
            "1646bdb0cfcd4ed75929fa1d96b17581",
        ),
        (
            (Klass(), ""),
            None,
            "{arg1}",
            "hoho",
        ),
        (
            ("a1",),
            {"arg2": 2, "kwarg1": "k1"},
            "{arg2}-{kwarg1}-{kwarg3}",
            "2-k1-",
        ),
        (
            ("a1",),
            {"arg2": 2, "kwarg1": "k1", "kwarg3": "k3"},
            "{arg2}:{kwarg1}:{kwarg3}",
            "2:k1:k3",
        ),
        (
            ("a1",),
            {"kwarg1": "K1", "arg2": 2},
            "{arg2}:{kwarg1}:{kwarg3}",
            "2:K1:",
        ),
        (("a1", "a2"), {"kwarg1": "test"}, "{kwarg1:len}", "4"),
        (
            ("a1", "a2"),
            {"user": type("user", (), {"name": "test"})()},
            "{user.name:len}",
            "4",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "test"},
            "{kwarg1:hash}",
            "098f6bcd4621d373cade4e832627b4f6",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "test"},
            "{kwarg1:hash(md5)}",
            "098f6bcd4621d373cade4e832627b4f6",
        ),
        (
            ("a1", "a2"),
            {"kwarg1": "eyJhbGciOiJIUzI1NiJ9.eyJ1c2VyIjoidGVzdCJ9.n75bSCIEuSemX5iop0sCF3HmXwMQWF1zI6TzckzHUKY"},
            "{kwarg1:jwt(user)}",
            "test",
        ),
    ),
)
def test_cache_key_args_kwargs(args, kwargs, template, key):
    async def func(arg1, arg2, *args, kwarg1=None, kwarg2=b"true", **kwargs):
        ...

    assert get_cache_key(func, template, args=args, kwargs=kwargs) == key


@pytest.mark.parametrize(
    ("func", "key", "template"),
    (
        (func1, None, "tests.test_key:func1:a:{a}"),
        (func2, None, "tests.test_key:func2:a:{a}:k:{k}:{__kwargs__}"),
        (func3, None, "tests.test_key:func3:a:{a}:k:{k}"),
        (Klass.method, None, "tests.test_key:Klass.method:self:{self}:a:{a}:k:{k}"),
        (Klass.method, "key:{k}:{self.data.test}", "key:{k}:{self.data.test}"),
        (func2, "key:{k}", "key:{k}"),
        (func3, "key:{k:len}:{k:hash(md5)}", "key:{k:len}:{k:hash(md5)}"),
    ),
)
def test_get_key_template(func, key, template):
    assert get_cache_key_template(func, key) == template


def test_get_key_template_error():
    with pytest.raises(ValueError) as exc:
        get_cache_key_template(func1, "key:{wrong_key}:{a}")
    exc.match("wrong_key")


@pytest.mark.parametrize(
    ("ttl", "expect"),
    (
        (timedelta(seconds=10), 10),
        (lambda: timedelta(hours=1), 60 * 60),
        (10, 10),
        (100.1, 100.1),
        ("10s", 10),
        ("1m10s", 60 + 10),
        ("10m1s", 60 * 10 + 1),
        ("1", 1),
        ("80", 80),
        (lambda: "1h", 60 * 60),
    ),
)
def test_ttl_to_seconds(ttl, expect):
    assert ttl_to_seconds(ttl) == expect



from functools import wraps
from typing import Optional

from ..._typing import CacheCondition
from ...backends.interface import Backend
from ...formatter import register_template
from ...key import get_cache_key, get_cache_key_template
from .defaults import CacheDetect, _empty, _get_cache_condition, context_cache_detect

__all__ = ("cache",)


def cache(
    backend: Backend,
    ttl: int,
    key: Optional[str] = None,
    condition: CacheCondition = None,
    prefix: str = "",
):
    """
    Simple cache strategy - trying to return cached result,
    execute wrapped call and store a result with ttl if condition return true
    :param backend: cache backend
    :param ttl: duration in seconds to store a result
    :param key: custom cache key, may contain alias to args or kwargs passed to a call
    :param condition: callable object that determines whether the result will be saved or not
    :param prefix: custom prefix for key
    """
    store = _get_cache_condition(condition)

    def _decor(func):
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        register_template(func, _key_template)

        @wraps(func)
        async def _wrap(*args, _from_cache: CacheDetect = context_cache_detect, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)
            cached = await backend.get(_cache_key, default=_empty)
            if cached is not _empty:
                _from_cache._set(_cache_key, ttl=ttl, name="simple", template=_key_template)
                return cached
            result = await func(*args, **kwargs)
            if store(result, args, kwargs, key=_cache_key):
                await backend.set(_cache_key, result, expire=ttl)
            return result

        return _wrap

    return _decor



import asyncio
from contextvars import ContextVar
from functools import wraps
from typing import Any, Callable, Dict, Optional, Union

from .backends.interface import Backend
from .formatter import get_templates_for_func, template_to_pattern
from .key import get_call_values, get_func_params


async def invalidate_func(backend: Backend, func, kwargs: Optional[Dict] = None) -> None:
    values = {**{param: "*" for param in get_func_params(func)}, **kwargs}
    for template in get_templates_for_func(func):
        del_template = template_to_pattern(template, **values)
        await backend.delete_match(del_template)


def invalidate(
    backend: Backend,
    target: Union[str, Callable],
    args_map: Optional[Dict[str, str]] = None,
    defaults: Optional[Dict[str, Any]] = None,
):
    args_map = args_map or {}
    defaults = defaults or {}

    def _decor(func):
        @wraps(func)
        async def _wrap(*args, **kwargs):
            result = await func(*args, **kwargs)
            _args = get_call_values(func, args, kwargs)
            _args.update(defaults)
            for source, dest in args_map.items():
                if dest in _args:
                    _args[source] = _args.pop(dest)
            if callable(target):
                asyncio.create_task(invalidate_func(backend, target, _args))
            else:
                key = target.format(**{k: str(v) if v is not None else "" for k, v in _args.items()})
                asyncio.create_task(backend.delete_match(key))
            return result

        return _wrap

    return _decor


_INVALIDATE_FURTHER = ContextVar("invalidate", default=False)


def set_invalidate_further():
    _INVALIDATE_FURTHER.set(True)


async def _invalidate_middleware(call, *args, key=None, backend=None, cmd=None, **kwargs):
    if _INVALIDATE_FURTHER.get() and key is not None and cmd != "delete":
        asyncio.create_task(backend.delete(key))
        _INVALIDATE_FURTHER.set(False)
        return None
    if key is None:
        return await call(*args, **kwargs)
    return await call(*args, key=key, **kwargs)



import logging
from functools import wraps
from typing import Any, Callable, NoReturn, Optional

from ..backends.interface import Backend
from ..formatter import register_template
from ..key import get_cache_key, get_cache_key_template

logger = logging.getLogger(__name__)


class RateLimitException(Exception):
    pass


def _default_action(*args: Any, **kwargs: Any) -> NoReturn:
    raise RateLimitException()


def rate_limit(
    backend: Backend,
    limit: int,
    period: int,
    ttl: int = None,
    key: Optional[str] = None,
    action: Optional[Callable] = None,
    prefix: str = "rate_limit",
):  # pylint: disable=too-many-arguments
    """
    Rate limit for function call. Do not call function if rate limit is reached, and call given action

    :param backend: cache backend
    :param limit: number of calls
    :param period: Period
    :param ttl: time ban, default == period
    :param action: call when rate limit reached, default raise RateLimitException
    :param prefix: custom prefix for key, default 'rate_limit'
    """
    action = _default_action if action is None else action

    def decorator(func):
        _key_template = get_cache_key_template(func, key=key, prefix=prefix)
        register_template(func, _key_template)

        @wraps(func)
        async def wrapped_func(*args, **kwargs):
            _cache_key = get_cache_key(func, _key_template, args, kwargs)

            requests_count = await backend.incr(key=_cache_key)  # set 1 if not exists
            if requests_count and requests_count > limit:
                if ttl and requests_count == limit + 1:
                    await backend.expire(key=_cache_key, timeout=ttl)
                logger.info("Rate limit reach for %s", _cache_key)
                action(*args, **kwargs)

            if requests_count == 1:
                await backend.expire(key=_cache_key, timeout=period)

            return await func(*args, **kwargs)

        return wrapped_func

    return decorator



from .backends.interface import LockedException  # noqa
from .decorators import (  # noqa
    NOT_NONE,
    CacheDetect,
    CircuitBreakerOpen,
    RateLimitException,
    context_cache_detect,
    fast_condition,
)
from .formatter import default_formatter, get_template_and_func_for, get_template_for_key  # noqa
from .helpers import add_prefix, all_keys_lower  # noqa
from .key import noself  # noqa
from .validation import set_invalidate_further  # noqa
from .wrapper import Cache  # noqa

cache_detect = context_cache_detect
# pylint: disable=invalid-name
cache = Cache(name="default")
noself_cache = noself(cache.cache)
rate_limit = cache.rate_limit
failover = cache.failover
circuit_breaker = cache.circuit_breaker
early = cache.early
soft = cache.soft
hit = cache.hit
dynamic = cache.dynamic
locked = cache.locked
invalidate = cache.invalidate
invalidate_func = cache.invalidate_func


mem = Cache(name="mem")
mem.setup(
    "mem://?check_interval=1",
    size=1_000_000,
)  # 1_000_000 * 248(size of small dict) == 31 mb



class Bitarray:
    __slots__ = ("_value",)

    def __init__(self, value, base=10):
        self._value = int(value, base=base)

    def get(self, index: int, size: int = 1) -> int:
        res = 0
        bit_index = 0
        for i in range(index * size, (index + 1) * size):
            res |= ((self._value >> i) & 1) << bit_index
            bit_index += 1
        return res

    def set(self, index: int, value: int, size: int = 1):
        for i in range(0, size):
            if (value >> i) & 1:
                self._set_bit_1(index * size + i)
            else:
                self._set_bit_0(index * size + i)
        return

    def _set_bit_1(self, index):
        self._value |= 1 << index

    def _set_bit_0(self, index):
        self._value &= ~(1 << index)

    def incr(self, index: int, size: int = 1, by: int = 1):
        if by > 0:
            by = min(by, 2**size - 1)
        else:
            by = max(by, -(2**size) - 1)
        value = self.get(index, size)
        value += by
        value = min(max(0, value), 2**size - 1)
        self.set(index, value, size)
        return self

    def copy(self):
        return Bitarray(str(self), 2)

    def to_int(self):
        return self._value

    def __str__(self):
        return f"{self._value:b}"



import asyncio
import sys

import pytest

from cashews.backends.interface import Backend
from cashews.backends.memory import Memory

pytestmark = pytest.mark.asyncio


@pytest.fixture(
    name="cache",
    params=[
        "memory",
        pytest.param("redis", marks=pytest.mark.redis),
        pytest.param("diskcache", marks=pytest.mark.diskcache),
    ],
)
async def _cache(request, redis_dsn, backend_factory):
    if request.param == "diskcache":
        from cashews.backends.diskcache import DiskCache

        backend = await backend_factory(DiskCache, shards=0)
        yield backend
    elif request.param == "redis":
        from cashews.backends.redis import Redis

        yield await backend_factory(Redis, redis_dsn, hash_key=None)
    else:
        yield await backend_factory(Memory)


async def test_set_get(cache):
    await cache.set("key", b"value")
    assert await cache.get("key") == b"value"


async def test_set_get_many(cache):
    await cache.set("key", b"value")
    assert await cache.get_many("key", "no_exists") == (b"value", None)


async def test_set_exist(cache):
    assert await cache.set("key", b"value")
    assert await cache.set("key", b"value", exist=True)
    assert not await cache.set("key2", b"value", exist=True)

    assert await cache.set("key2", b"value", exist=False)
    assert not await cache.set("key2", b"value", exist=False)


async def test_get_no_value(cache):
    assert await cache.get("key2") is None


async def test_incr(cache):
    assert await cache.incr("incr") == 1
    assert await cache.incr("incr") == 2
    assert await cache.get("incr") == 2


async def test_incr_setted(cache):
    await cache.set("incr", "test")
    with pytest.raises(Exception):
        assert await cache.incr("incr") == 1
    assert await cache.get("incr") == "test"


async def test_ping(cache):
    assert await cache.ping() == b"PONG"


async def test_exists(cache):
    assert not await cache.exists("not")
    await cache.set("yes", "1")
    assert await cache.exists("yes")


async def test_expire(cache):
    await cache.set("key", b"value", expire=0.01)
    assert await cache.get("key") == b"value"
    await asyncio.sleep(0.011)
    assert await cache.get("key") is None


async def test_get_set_expire(cache):
    await cache.set("key", b"value")
    assert await cache.get("key") == b"value"
    assert await cache.get_expire("key") == -1
    await cache.expire("key", 1)
    assert await cache.get_expire("key") == 1


async def test_delete_match(cache: Backend):
    await cache.set("pref:test:test", b"value")
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    await cache.delete_match("pref:*:test")

    assert await cache.get("pref:test:test") is None
    assert await cache.get("pref:value:test") is None
    assert await cache.get("pref:-:test") is None
    assert await cache.get("pref:*:test") is None

    assert await cache.get("ppref:test:test") is not None
    assert await cache.get("pref:test:tests") is not None


async def test_scan(cache: Backend):
    await cache.set("pref:test:test", b"value")
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    keys = [key async for key in cache.scan("pref:*:test")]

    assert len(keys) == 4
    assert set(keys) == {"pref:test:test", "pref:value:test", "pref:-:test", "pref:*:test"}


async def test_get_match(cache: Backend):
    await cache.set("pref:test:test", b"value")
    await cache.set("pref:value:test", b"value2")
    await cache.set("pref:-:test", b"-")
    await cache.set("pref:*:test", b"*")

    await cache.set("ppref:test:test", b"value3")
    await cache.set("pref:test:tests", b"value3")

    match = [(key, value) async for key, value in cache.get_match("pref:*:test")]

    assert len(match) == 4
    assert dict(match) == {
        "pref:test:test": b"value",
        "pref:value:test": b"value2",
        "pref:-:test": b"-",
        "pref:*:test": b"*",
    }
    match = [(key, value) async for key, value in cache.get_match("not_exists:*")]
    assert len(match) == 0


async def test_get_size(cache: Backend):
    await cache.set("test", b"1")
    assert await cache.get_size("test") in (
        sys.getsizeof((None, b"1")) + sys.getsizeof(b"1") + sys.getsizeof(None),  # ordered dict
        66,  # redis 6
        72,  # redis 7
        -1,  # diskcache
    )


async def test_get_bits(cache: Backend):
    assert await cache.get_bits("test", 0, 2, 10, 50000, size=1) == (0, 0, 0, 0)
    assert await cache.get_bits("test", 0, 1, 3, size=15) == (
        0,
        0,
        0,
    )


async def test_incr_bits(cache: Backend):
    await cache.incr_bits("test", 0, 1, 4)
    assert await cache.get_bits("test", 0, 1, 2, 3, 4) == (1, 1, 0, 0, 1)


async def test_bits_size(cache: Backend):
    await cache.incr_bits("test", 0, 1, 4, size=5, by=3)
    assert await cache.get_bits("test", 0, 1, 2, 3, 4, size=5) == (3, 3, 0, 0, 3)


async def test_lru(backend_factory):
    cache = await backend_factory(Memory, size=10)
    # fill cache
    for i in range(10):
        await cache.set(f"key:{i}", i)

    # use only 5 first
    for i in range(5):
        await cache.get(f"key:{i}")

    # add 5 more keys
    for i in range(5):
        await cache.set(f"key:{i}:new", i)

    assert len(cache.store) == 10

    for i in range(5):
        assert await cache.get(f"key:{i}") == i

    for i in range(6, 10):
        assert await cache.get(f"key:{i}") == None


async def test_lru2(backend_factory):
    cache = await backend_factory(Memory, size=10)
    # fill cache
    for i in range(10):
        await cache.set(f"key:{i}", i)

    # use only 5 last
    for i in range(6, 10):
        await cache.get(f"key:{i}")

    # add 5 more keys
    for i in range(5):
        await cache.set(f"key:{i}:new", i)

    assert len(cache.store) == 10

    for i in range(5):
        assert await cache.get(f"key:{i}") == None

    for i in range(6, 10):
        assert await cache.get(f"key:{i}") == i
